"""This module provides abstract base classes for CAN message reading and writing operations
to various file formats.

.. note::
    All classes in this module are abstract and should be subclassed to implement
    specific file format handling.
"""

import locale
import os
from abc import ABC, abstractmethod
from collections.abc import Iterable
from contextlib import AbstractContextManager
from io import BufferedIOBase, TextIOWrapper
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Generic,
    Literal,
    Optional,
    TextIO,
    TypeVar,
    Union,
)

from typing_extensions import Self

from ..listener import Listener
from ..message import Message
from ..typechecking import FileLike, StringPathLike

if TYPE_CHECKING:
    from _typeshed import (
        OpenBinaryModeReading,
        OpenBinaryModeUpdating,
        OpenBinaryModeWriting,
        OpenTextModeReading,
        OpenTextModeUpdating,
        OpenTextModeWriting,
    )


#: type parameter used in generic classes :class:`MessageReader` and :class:`MessageWriter`
_IoTypeVar = TypeVar("_IoTypeVar", bound=FileLike)


class MessageWriter(AbstractContextManager["MessageWriter"], Listener, ABC):
    """Abstract base class for all CAN message writers.

    This class serves as a foundation for implementing different message writer formats.
    It combines context manager capabilities with the message listener interface.

    :param file: Path-like object or string representing the output file location
    :param kwargs: Additional keyword arguments for specific writer implementations
    """

    @abstractmethod
    def __init__(self, file: StringPathLike, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        """Stop handling messages and cleanup any resources."""

    def __enter__(self) -> Self:
        """Enter the context manager."""
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        """Exit the context manager and ensure proper cleanup."""
        self.stop()
        return False


class SizedMessageWriter(MessageWriter, ABC):
    """Abstract base class for message writers that can report their file size.

    This class extends :class:`MessageWriter` with the ability to determine the size
    of the output file.
    """

    @abstractmethod
    def file_size(self) -> int:
        """Get the current size of the output file in bytes.

        :return: The size of the file in bytes
        :rtype: int
        """


class FileIOMessageWriter(SizedMessageWriter, Generic[_IoTypeVar]):
    """Base class for writers that operate on file descriptors.

    This class provides common functionality for writers that work with file objects.

    :param file: A path-like object or file object to write to
    :param kwargs: Additional keyword arguments for specific writer implementations

    :ivar file: The file object being written to
    """

    file: _IoTypeVar

    @abstractmethod
    def __init__(self, file: Union[StringPathLike, _IoTypeVar], **kwargs: Any) -> None:
        pass

    def stop(self) -> None:
        """Close the file and stop writing."""
        self.file.close()

    def file_size(self) -> int:
        """Get the current file size."""
        return self.file.tell()


class TextIOMessageWriter(FileIOMessageWriter[Union[TextIO, TextIOWrapper]], ABC):
    """Text-based message writer implementation.

    :param file: Text file to write to
    :param mode: File open mode for text operations
    :param kwargs: Additional arguments like encoding
    """

    def __init__(
        self,
        file: Union[StringPathLike, TextIO, TextIOWrapper],
        mode: "Union[OpenTextModeUpdating, OpenTextModeWriting]" = "w",
        **kwargs: Any,
    ) -> None:
        if isinstance(file, (str, os.PathLike)):
            encoding: str = kwargs.get("encoding", locale.getpreferredencoding(False))
            # pylint: disable=consider-using-with
            self.file = Path(file).open(mode=mode, encoding=encoding)
        else:
            self.file = file


class BinaryIOMessageWriter(FileIOMessageWriter[Union[BinaryIO, BufferedIOBase]], ABC):
    """Binary file message writer implementation.

    :param file: Binary file to write to
    :param mode: File open mode for binary operations
    :param kwargs: Additional implementation specific arguments
    """

    def __init__(
        self,
        file: Union[StringPathLike, BinaryIO, BufferedIOBase],
        mode: "Union[OpenBinaryModeUpdating, OpenBinaryModeWriting]" = "wb",
        **kwargs: Any,
    ) -> None:
        if isinstance(file, (str, os.PathLike)):
            # pylint: disable=consider-using-with,unspecified-encoding
            self.file = Path(file).open(mode=mode)
        else:
            self.file = file


class MessageReader(AbstractContextManager["MessageReader"], Iterable[Message], ABC):
    """Abstract base class for all CAN message readers.

    This class serves as a foundation for implementing different message reader formats.
    It combines context manager capabilities with iteration interface.

    :param file: Path-like object or string representing the input file location
    :param kwargs: Additional keyword arguments for specific reader implementations
    """

    @abstractmethod
    def __init__(self, file: StringPathLike, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        """Stop reading messages and cleanup any resources."""

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        self.stop()
        return False


class FileIOMessageReader(MessageReader, Generic[_IoTypeVar]):
    """Base class for readers that operate on file descriptors.

    This class provides common functionality for readers that work with file objects.

    :param file: A path-like object or file object to read from
    :param kwargs: Additional keyword arguments for specific reader implementations

    :ivar file: The file object being read from
    """

    file: _IoTypeVar

    @abstractmethod
    def __init__(self, file: Union[StringPathLike, _IoTypeVar], **kwargs: Any) -> None:
        pass

    def stop(self) -> None:
        self.file.close()


class TextIOMessageReader(FileIOMessageReader[Union[TextIO, TextIOWrapper]], ABC):
    """Text-based message reader implementation.

    :param file: Text file to read from
    :param mode: File open mode for text operations
    :param kwargs: Additional arguments like encoding
    """

    def __init__(
        self,
        file: Union[StringPathLike, TextIO, TextIOWrapper],
        mode: "OpenTextModeReading" = "r",
        **kwargs: Any,
    ) -> None:
        if isinstance(file, (str, os.PathLike)):
            encoding: str = kwargs.get("encoding", locale.getpreferredencoding(False))
            # pylint: disable=consider-using-with
            self.file = Path(file).open(mode=mode, encoding=encoding)
        else:
            self.file = file


class BinaryIOMessageReader(FileIOMessageReader[Union[BinaryIO, BufferedIOBase]], ABC):
    """Binary file message reader implementation.

    :param file: Binary file to read from
    :param mode: File open mode for binary operations
    :param kwargs: Additional implementation specific arguments
    """

    def __init__(
        self,
        file: Union[StringPathLike, BinaryIO, BufferedIOBase],
        mode: "OpenBinaryModeReading" = "rb",
        **kwargs: Any,
    ) -> None:
        if isinstance(file, (str, os.PathLike)):
            # pylint: disable=consider-using-with,unspecified-encoding
            self.file = Path(file).open(mode=mode)
        else:
            self.file = file



import errno
import logging
import platform
import select
import socket
import struct
import time
import warnings
from typing import Any, Optional, Union

import can
from can import BusABC, CanProtocol, Message
from can.typechecking import AutoDetectedConfig

from .utils import is_msgpack_installed, pack_message, unpack_message

is_linux = platform.system() == "Linux"
if is_linux:
    from fcntl import ioctl

log = logging.getLogger(__name__)


# see socket.getaddrinfo()
IPv4_ADDRESS_INFO = tuple[str, int]  # address, port
IPv6_ADDRESS_INFO = tuple[str, int, int, int]  # address, port, flowinfo, scope_id
IP_ADDRESS_INFO = Union[IPv4_ADDRESS_INFO, IPv6_ADDRESS_INFO]

# Additional constants for the interaction with Unix kernels
SO_TIMESTAMPNS = 35
SIOCGSTAMP = 0x8906

# Additional constants for the interaction with the Winsock API
WSAEINVAL = 10022


class UdpMulticastBus(BusABC):
    """A virtual interface for CAN communications between multiple processes using UDP over Multicast IP.

    It supports IPv4 and IPv6, specified via the channel (which really is just a multicast IP address as a
    string). You can also specify the port and the IPv6 *hop limit*/the IPv4 *time to live* (TTL).

    This bus does not support filtering based on message IDs on the kernel level but instead provides it in
    user space (in Python) as a fallback.

    Both default addresses should allow for multi-host CAN networks in a normal local area network (LAN) where
    multicast is enabled.

    .. note::
        The auto-detection of available interfaces (see) is implemented using heuristic that checks if the
        required socket operations are available. It then returns two configurations, one based on
        the :attr:`~UdpMulticastBus.DEFAULT_GROUP_IPv6` address and another one based on
        the :attr:`~UdpMulticastBus.DEFAULT_GROUP_IPv4` address.

    .. warning::
        The parameter `receive_own_messages` is currently unsupported and setting it to `True` will raise an
        exception.

    .. warning::
        This interface does not make guarantees on reliable delivery and message ordering, and also does not
        implement rate limiting or ID arbitration/prioritization under high loads. Please refer to the section
        :ref:`virtual_interfaces_doc` for more information on this and a comparison to alternatives.

    :param channel: A multicast IPv4 address (in `224.0.0.0/4`) or an IPv6 address (in `ff00::/8`).
                    This defines which version of IP is used. See
                    `Wikipedia ("Multicast address") <https://en.wikipedia.org/wiki/Multicast_address>`__
                    for more details on the addressing schemes.
                    Defaults to :attr:`~UdpMulticastBus.DEFAULT_GROUP_IPv6`.
    :param port: The IP port to read from and write to.
    :param hop_limit: The hop limit in IPv6 or in IPv4 the time to live (TTL).
    :param receive_own_messages: If transmitted messages should also be received by this bus.
                                 CURRENTLY UNSUPPORTED.
    :param fd:
        If CAN-FD frames should be supported. If set to false, an error will be raised upon sending such a
        frame and such received frames will be ignored.
    :param can_filters: See :meth:`~can.BusABC.set_filters`.

    :raises RuntimeError: If the *msgpack*-dependency is not available. It should be installed on all
                          non Windows platforms via the `setup.py` requirements.
    :raises NotImplementedError: If the `receive_own_messages` is passed as `True`.
    """

    #: An arbitrary IPv6 multicast address with "site-local" scope, i.e. only to be routed within the local
    #: physical network and not beyond it. It should allow for multi-host CAN networks in a normal IPv6 LAN.
    #: This is the default channel and should work with most modern routers if multicast is allowed.
    DEFAULT_GROUP_IPv6 = "ff15:7079:7468:6f6e:6465:6d6f:6d63:6173"

    #: An arbitrary IPv4 multicast address with "administrative" scope, i.e. only to be routed within
    #: administrative organizational boundaries and not beyond it.
    #: It should allow for multi-host CAN networks in a normal IPv4 LAN.
    #: This is provided as a default fallback channel if IPv6 is (still) not supported.
    DEFAULT_GROUP_IPv4 = "239.74.163.2"

    def __init__(
        self,
        channel: str = DEFAULT_GROUP_IPv6,
        port: int = 43113,
        hop_limit: int = 1,
        receive_own_messages: bool = False,
        fd: bool = True,
        **kwargs: Any,
    ) -> None:
        is_msgpack_installed()

        if receive_own_messages:
            raise can.CanInterfaceNotImplementedError(
                "receiving own messages is not yet implemented"
            )

        super().__init__(
            channel,
            **kwargs,
        )

        self._multicast = GeneralPurposeUdpMulticastBus(channel, port, hop_limit)
        self._can_protocol = CanProtocol.CAN_FD if fd else CanProtocol.CAN_20

    @property
    def is_fd(self) -> bool:
        class_name = self.__class__.__name__
        warnings.warn(
            f"The {class_name}.is_fd property is deprecated and superseded by "
            f"{class_name}.protocol. It is scheduled for removal in python-can version 5.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._can_protocol is CanProtocol.CAN_FD

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> tuple[Optional[Message], bool]:
        result = self._multicast.recv(timeout)
        if not result:
            return None, False

        data, _, timestamp = result
        try:
            can_message = unpack_message(
                data, replace={"timestamp": timestamp}, check=True
            )
        except Exception as exception:
            raise can.CanOperationError(
                "could not unpack received message"
            ) from exception

        if self._can_protocol is not CanProtocol.CAN_FD and can_message.is_fd:
            return None, False

        return can_message, False

    def send(self, msg: can.Message, timeout: Optional[float] = None) -> None:
        if self._can_protocol is not CanProtocol.CAN_FD and msg.is_fd:
            raise can.CanOperationError(
                "cannot send FD message over bus with CAN FD disabled"
            )

        data = pack_message(msg)
        self._multicast.send(data, timeout)

    def fileno(self) -> int:
        """Provides the internally used file descriptor of the socket or `-1` if not available."""
        return self._multicast.fileno()

    def shutdown(self) -> None:
        """Close all sockets and free up any resources.

        Never throws errors and only logs them.
        """
        super().shutdown()
        self._multicast.shutdown()

    @staticmethod
    def _detect_available_configs() -> list[AutoDetectedConfig]:
        if hasattr(socket, "CMSG_SPACE"):
            return [
                {
                    "interface": "udp_multicast",
                    "channel": UdpMulticastBus.DEFAULT_GROUP_IPv6,
                },
                {
                    "interface": "udp_multicast",
                    "channel": UdpMulticastBus.DEFAULT_GROUP_IPv4,
                },
            ]

        # else, this interface cannot be used
        return []


class GeneralPurposeUdpMulticastBus:
    """A general purpose send and receive handler for multicast over IP/UDP.

    However, it raises CAN-specific exceptions for convenience.
    """

    def __init__(
        self, group: str, port: int, hop_limit: int, max_buffer: int = 4096
    ) -> None:
        self.group = group
        self.port = port
        self.hop_limit = hop_limit
        self.max_buffer = max_buffer

        # `False` will always work, no matter the setup. This might be changed by _create_socket().
        self.timestamp_nanosecond = False

        # Look up multicast group address in name server and find out IP version of the first suitable target
        # and then get the address family of it (socket.AF_INET or socket.AF_INET6)
        connection_candidates = socket.getaddrinfo(
            group, self.port, type=socket.SOCK_DGRAM
        )
        sock = None
        for connection_candidate in connection_candidates:
            address_family: socket.AddressFamily = connection_candidate[0]
            self.ip_version = 4 if address_family == socket.AF_INET else 6
            try:
                sock = self._create_socket(address_family)
            except OSError as error:
                log.info(
                    "could not connect to the multicast IP network of candidate %s; reason: %s",
                    connection_candidates,
                    error,
                )
        if sock is not None:
            self._socket = sock
        else:
            raise can.CanInitializationError(
                "could not connect to a multicast IP network"
            )

        # used in recv()
        self.received_timestamp_struct = "@ll"
        self.received_timestamp_struct_size = struct.calcsize(
            self.received_timestamp_struct
        )
        if self.timestamp_nanosecond:
            self.received_ancillary_buffer_size = socket.CMSG_SPACE(
                self.received_timestamp_struct_size
            )
        else:
            self.received_ancillary_buffer_size = 0

        # used by send()
        self._send_destination = (self.group, self.port)
        self._last_send_timeout: Optional[float] = None

    def _create_socket(self, address_family: socket.AddressFamily) -> socket.socket:
        """Creates a new socket. This might fail and raise an exception!

        :param address_family: whether this is of type `socket.AF_INET` or `socket.AF_INET6`

        :raises can.CanInitializationError:
            if the socket could not be opened or configured correctly; in this case, it is
            guaranteed to be closed/cleaned up
        """
        # create the UDP socket
        # this might already fail but then there is nothing to clean up
        sock = socket.socket(address_family, socket.SOCK_DGRAM)

        # configure the socket
        try:
            # set hop limit / TTL
            ttl_as_binary = struct.pack("@I", self.hop_limit)
            if self.ip_version == 4:
                sock.setsockopt(
                    socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl_as_binary
                )
            else:
                sock.setsockopt(
                    socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, ttl_as_binary
                )

            # Allow multiple programs to access that address + port
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            # Option not supported on Windows.
            if hasattr(socket, "SO_REUSEPORT"):
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

            # set how to receive timestamps
            try:
                sock.setsockopt(socket.SOL_SOCKET, SO_TIMESTAMPNS, 1)
            except OSError as error:
                if (
                    error.errno == errno.ENOPROTOOPT
                    or error.errno == errno.EINVAL
                    or error.errno == WSAEINVAL
                ):  # It is unavailable on macOS (ENOPROTOOPT) or windows(EINVAL/WSAEINVAL)
                    self.timestamp_nanosecond = False
                else:
                    raise error
            else:
                self.timestamp_nanosecond = True

            # Bind it to the port (on any interface)
            sock.bind(("", self.port))

            # Join the multicast group
            group_as_binary = socket.inet_pton(address_family, self.group)
            if self.ip_version == 4:
                request = group_as_binary + struct.pack("@I", socket.INADDR_ANY)
                sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, request)
            else:
                request = group_as_binary + struct.pack("@I", 0)
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, request)

            return sock

        except OSError as error:
            # clean up the incompletely configured but opened socket
            try:
                sock.close()
            except OSError as close_error:
                # ignore but log any failures in here
                log.warning("Could not close partly configured socket: %s", close_error)

            # still raise the error
            raise can.CanInitializationError(
                "could not create or configure socket"
            ) from error

    def send(self, data: bytes, timeout: Optional[float] = None) -> None:
        """Send data to all group members. This call blocks.

        :param timeout: the timeout in seconds after which an Exception is raised is sending has failed
        :param data: the data to be sent
        :raises can.CanOperationError: if an error occurred while writing to the underlying socket
        :raises can.CanTimeoutError: if the timeout ran out before sending was completed
        """
        if timeout != self._last_send_timeout:
            self._last_send_timeout = timeout
            # this applies to all blocking calls on the socket, but sending is the only one that is blocking
            self._socket.settimeout(timeout)

        try:
            bytes_sent = self._socket.sendto(data, self._send_destination)
            if bytes_sent < len(data):
                raise TimeoutError()
        except TimeoutError:
            raise can.CanTimeoutError() from None
        except OSError as error:
            raise can.CanOperationError("failed to send via socket") from error

    def recv(
        self, timeout: Optional[float] = None
    ) -> Optional[tuple[bytes, IP_ADDRESS_INFO, float]]:
        """
        Receive up to **max_buffer** bytes.

        :param timeout: the timeout in seconds after which `None` is returned if no data arrived
        :returns: `None` on timeout, or a 3-tuple comprised of:
            - received data,
            - the sender of the data, and
            - a timestamp in seconds
        """
        # get all sockets that are ready (can be a list with a single value
        # being self.socket or an empty list if self.socket is not ready)
        try:
            # get all sockets that are ready (can be a list with a single value
            # being self.socket or an empty list if self.socket is not ready)
            ready_receive_sockets, _, _ = select.select([self._socket], [], [], timeout)
        except OSError as exc:
            # something bad (not a timeout) happened (e.g. the interface went down)
            raise can.CanOperationError(
                f"Failed to wait for IP/UDP socket: {exc}"
            ) from exc

        if ready_receive_sockets:  # not empty
            # fetch timestamp; this is configured in _create_socket()
            if self.timestamp_nanosecond:
                # fetch data, timestamp & source address
                (
                    raw_message_data,
                    ancillary_data,
                    _,  # flags
                    sender_address,
                ) = self._socket.recvmsg(
                    self.max_buffer, self.received_ancillary_buffer_size
                )

                # Very similar to timestamp handling in can/interfaces/socketcan/socketcan.py -> capture_message()
                if len(ancillary_data) != 1:
                    raise can.CanOperationError(
                        "Only requested a single extra field but got a different amount"
                    )
                cmsg_level, cmsg_type, cmsg_data = ancillary_data[0]
                if cmsg_level != socket.SOL_SOCKET or cmsg_type != SO_TIMESTAMPNS:
                    raise can.CanOperationError(
                        "received control message type that was not requested"
                    )
                # see https://man7.org/linux/man-pages/man3/timespec.3.html -> struct timespec for details
                seconds, nanoseconds = struct.unpack(
                    self.received_timestamp_struct, cmsg_data
                )
                if nanoseconds >= 1e9:
                    raise can.CanOperationError(
                        f"Timestamp nanoseconds field was out of range: {nanoseconds} not less than 1e9"
                    )
                timestamp = seconds + nanoseconds * 1.0e-9
            else:
                # fetch data & source address
                (raw_message_data, sender_address) = self._socket.recvfrom(
                    self.max_buffer
                )

                if is_linux:
                    # This ioctl isn't supported on Darwin & Windows.
                    result_buffer = ioctl(
                        self._socket.fileno(),
                        SIOCGSTAMP,
                        bytes(self.received_timestamp_struct_size),
                    )
                    seconds, microseconds = struct.unpack(
                        self.received_timestamp_struct, result_buffer
                    )
                else:
                    # fallback to time.time_ns
                    now = time.time()

                    # Extract seconds and microseconds
                    seconds = int(now)
                    microseconds = int((now - seconds) * 1000000)

                if microseconds >= 1e6:
                    raise can.CanOperationError(
                        f"Timestamp microseconds field was out of range: {microseconds} not less than 1e6"
                    )
                timestamp = seconds + microseconds * 1e-6

            return raw_message_data, sender_address, timestamp

        # socket wasn't readable or timeout occurred
        return None

    def fileno(self) -> int:
        """Provides the internally used file descriptor of the socket or `-1` if not available."""
        return self._socket.fileno()

    def shutdown(self) -> None:
        """Close all sockets and free up any resources.

        Never throws errors and only logs them.
        """
        try:
            self._socket.close()
        except OSError as exception:
            log.error("could not close IP socket: %s", exception)



import argparse
import errno
import sys
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Union,
)

from can import BusState, Logger, SizedRotatingLogger
from can.cli import (
    _add_extra_args,
    _parse_additional_config,
    _set_logging_level_from_namespace,
    add_bus_arguments,
    create_bus_from_namespace,
)
from can.typechecking import TAdditionalCliArgs

if TYPE_CHECKING:
    from can.io import BaseRotatingLogger
    from can.io.generic import MessageWriter


def _parse_logger_args(
    args: list[str],
) -> tuple[argparse.Namespace, TAdditionalCliArgs]:
    """Parse command line arguments for logger script."""

    parser = argparse.ArgumentParser(
        description="Log CAN traffic, printing messages to stdout or to a "
        "given file.",
    )

    logger_group = parser.add_argument_group("logger arguments")

    logger_group.add_argument(
        "-f",
        "--file_name",
        dest="log_file",
        help="Path and base log filename, for supported types see can.Logger.",
        default=None,
    )

    logger_group.add_argument(
        "-a",
        "--append",
        dest="append",
        help="Append to the log file if it already exists.",
        action="store_true",
    )

    logger_group.add_argument(
        "-s",
        "--file_size",
        dest="file_size",
        type=int,
        help="Maximum file size in bytes. Rotate log file when size threshold "
        "is reached. (The resulting file sizes will be consistent, but are not "
        "guaranteed to be exactly what is specified here due to the rollover "
        "conditions being logger implementation specific.)",
        default=None,
    )

    logger_group.add_argument(
        "-v",
        action="count",
        dest="verbosity",
        help="""How much information do you want to see at the command line?
                        You can add several of these e.g., -vv is DEBUG""",
        default=2,
    )

    state_group = logger_group.add_mutually_exclusive_group(required=False)
    state_group.add_argument(
        "--active",
        help="Start the bus as active, this is applied by default.",
        action="store_true",
    )
    state_group.add_argument(
        "--passive", help="Start the bus as passive.", action="store_true"
    )

    # handle remaining arguments
    _add_extra_args(logger_group)

    # add bus options
    add_bus_arguments(parser, filter_arg=True)

    # print help message when no arguments were given
    if not args:
        parser.print_help(sys.stderr)
        raise SystemExit(errno.EINVAL)

    results, unknown_args = parser.parse_known_args(args)
    additional_config = _parse_additional_config([*results.extra_args, *unknown_args])
    return results, additional_config


def main() -> None:
    results, additional_config = _parse_logger_args(sys.argv[1:])
    bus = create_bus_from_namespace(results)
    _set_logging_level_from_namespace(results)

    if results.active:
        bus.state = BusState.ACTIVE
    elif results.passive:
        bus.state = BusState.PASSIVE

    print(f"Connected to {bus.__class__.__name__}: {bus.channel_info}")
    print(f"Can Logger (Started on {datetime.now()})")

    logger: Union[MessageWriter, BaseRotatingLogger]
    if results.file_size:
        logger = SizedRotatingLogger(
            base_filename=results.log_file,
            max_bytes=results.file_size,
            append=results.append,
            **additional_config,
        )
    else:
        logger = Logger(
            filename=results.log_file,
            append=results.append,
            **additional_config,
        )

    try:
        while True:
            msg = bus.recv(1)
            if msg is not None:
                logger(msg)
    except KeyboardInterrupt:
        pass
    finally:
        bus.shutdown()
        logger.stop()


if __name__ == "__main__":
    main()



"""
Contains handling of MF4 logging files.

MF4 files represent Measurement Data Format (MDF) version 4 as specified by
the ASAM MDF standard (see https://www.asam.net/standards/detail/mdf/)
"""

import abc
import heapq
import logging
from collections.abc import Generator, Iterator
from datetime import datetime
from hashlib import md5
from io import BufferedIOBase, BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Optional, Union, cast

from ..message import Message
from ..typechecking import StringPathLike
from ..util import channel2int, len2dlc
from .generic import BinaryIOMessageReader, BinaryIOMessageWriter

logger = logging.getLogger("can.io.mf4")

try:
    import asammdf
    import numpy as np
    from asammdf import Signal, Source
    from asammdf.blocks.mdf_v4 import MDF4
    from asammdf.blocks.v4_blocks import ChannelGroup, SourceInformation
    from asammdf.blocks.v4_constants import BUS_TYPE_CAN, FLAG_CG_BUS_EVENT, SOURCE_BUS
    from asammdf.mdf import MDF

    STD_DTYPE = np.dtype(
        [
            ("CAN_DataFrame.BusChannel", "<u1"),
            ("CAN_DataFrame.ID", "<u4"),
            ("CAN_DataFrame.IDE", "<u1"),
            ("CAN_DataFrame.DLC", "<u1"),
            ("CAN_DataFrame.DataLength", "<u1"),
            ("CAN_DataFrame.DataBytes", "(64,)u1"),
            ("CAN_DataFrame.Dir", "<u1"),
            ("CAN_DataFrame.EDL", "<u1"),
            ("CAN_DataFrame.BRS", "<u1"),
            ("CAN_DataFrame.ESI", "<u1"),
        ]
    )

    ERR_DTYPE = np.dtype(
        [
            ("CAN_ErrorFrame.BusChannel", "<u1"),
            ("CAN_ErrorFrame.ID", "<u4"),
            ("CAN_ErrorFrame.IDE", "<u1"),
            ("CAN_ErrorFrame.DLC", "<u1"),
            ("CAN_ErrorFrame.DataLength", "<u1"),
            ("CAN_ErrorFrame.DataBytes", "(64,)u1"),
            ("CAN_ErrorFrame.Dir", "<u1"),
            ("CAN_ErrorFrame.EDL", "<u1"),
            ("CAN_ErrorFrame.BRS", "<u1"),
            ("CAN_ErrorFrame.ESI", "<u1"),
        ]
    )

    RTR_DTYPE = np.dtype(
        [
            ("CAN_RemoteFrame.BusChannel", "<u1"),
            ("CAN_RemoteFrame.ID", "<u4"),
            ("CAN_RemoteFrame.IDE", "<u1"),
            ("CAN_RemoteFrame.DLC", "<u1"),
            ("CAN_RemoteFrame.DataLength", "<u1"),
            ("CAN_RemoteFrame.Dir", "<u1"),
        ]
    )
except ImportError:
    asammdf = None
    MDF4 = None
    Signal = None


CAN_MSG_EXT = 0x80000000
CAN_ID_MASK = 0x1FFFFFFF


class MF4Writer(BinaryIOMessageWriter):
    """Logs CAN data to an ASAM Measurement Data File v4 (.mf4).

    MF4Writer does not support append mode.

    If a message has a timestamp smaller than the previous one or None,
    it gets assigned the timestamp that was written for the last message.
    It the first message does not have a timestamp, it is set to zero.
    """

    def __init__(
        self,
        file: Union[StringPathLike, BinaryIO],
        database: Optional[StringPathLike] = None,
        compression_level: int = 2,
        **kwargs: Any,
    ) -> None:
        """
        :param file:
            A path-like object or as file-like object to write to.
            If this is a file-like object, is has to be opened in
            binary write mode, not text write mode.
        :param database:
            optional path to a DBC or ARXML file that contains message description.
        :param compression_level:
            compression option as integer (default 2)
            * 0 - no compression
            * 1 - deflate (slower, but produces smaller files)
            * 2 - transposition + deflate (slowest, but produces the smallest files)
        """
        if asammdf is None:
            raise NotImplementedError(
                "The asammdf package was not found. Install python-can with "
                "the optional dependency [mf4] to use the MF4Writer."
            )

        if kwargs.get("append", False):
            raise ValueError(
                f"{self.__class__.__name__} is currently not equipped to "
                f"append messages to an existing file."
            )

        super().__init__(file, mode="w+b")
        now = datetime.now()
        self._mdf = cast("MDF4", MDF(version="4.10"))
        self._mdf.header.start_time = now
        self.last_timestamp = self._start_time = now.timestamp()

        self._compression_level = compression_level

        if database:
            database = Path(database).resolve()
            if database.exists():
                data = database.read_bytes()
                attachment = data, database.name, md5(data).digest()
            else:
                attachment = None
        else:
            attachment = None

        acquisition_source = SourceInformation(
            source_type=SOURCE_BUS, bus_type=BUS_TYPE_CAN
        )

        # standard frames group
        self._mdf.append(
            Signal(
                name="CAN_DataFrame",
                samples=np.array([], dtype=STD_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        # error frames group
        self._mdf.append(
            Signal(
                name="CAN_ErrorFrame",
                samples=np.array([], dtype=ERR_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        # remote frames group
        self._mdf.append(
            Signal(
                name="CAN_RemoteFrame",
                samples=np.array([], dtype=RTR_DTYPE),
                timestamps=np.array([], dtype="<f8"),
                attachment=attachment,
                source=acquisition_source,
            )
        )

        self._std_buffer = np.zeros(1, dtype=STD_DTYPE)
        self._err_buffer = np.zeros(1, dtype=ERR_DTYPE)
        self._rtr_buffer = np.zeros(1, dtype=RTR_DTYPE)

    def file_size(self) -> int:
        """Return an estimate of the current file size in bytes."""
        # TODO: find solution without accessing private attributes of asammdf
        return cast(
            "int",
            self._mdf._tempfile.tell(),  # pylint: disable=protected-access,no-member
        )

    def stop(self) -> None:
        self._mdf.save(self.file, compression=self._compression_level)
        self._mdf.close()
        super().stop()

    def on_message_received(self, msg: Message) -> None:
        channel = channel2int(msg.channel)

        timestamp = msg.timestamp
        if timestamp is None:
            timestamp = self.last_timestamp
        else:
            self.last_timestamp = max(self.last_timestamp, timestamp)

        timestamp -= self._start_time

        if msg.is_remote_frame:
            if channel is not None:
                self._rtr_buffer["CAN_RemoteFrame.BusChannel"] = channel

            self._rtr_buffer["CAN_RemoteFrame.ID"] = msg.arbitration_id
            self._rtr_buffer["CAN_RemoteFrame.IDE"] = int(msg.is_extended_id)
            self._rtr_buffer["CAN_RemoteFrame.Dir"] = 0 if msg.is_rx else 1
            self._rtr_buffer["CAN_RemoteFrame.DLC"] = msg.dlc

            sigs = [(np.array([timestamp]), None), (self._rtr_buffer, None)]
            self._mdf.extend(2, sigs)

        elif msg.is_error_frame:
            if channel is not None:
                self._err_buffer["CAN_ErrorFrame.BusChannel"] = channel

            self._err_buffer["CAN_ErrorFrame.ID"] = msg.arbitration_id
            self._err_buffer["CAN_ErrorFrame.IDE"] = int(msg.is_extended_id)
            self._err_buffer["CAN_ErrorFrame.Dir"] = 0 if msg.is_rx else 1
            data = msg.data
            size = len(data)
            self._err_buffer["CAN_ErrorFrame.DataLength"] = size
            self._err_buffer["CAN_ErrorFrame.DataBytes"][0, :size] = data
            if msg.is_fd:
                self._err_buffer["CAN_ErrorFrame.DLC"] = len2dlc(msg.dlc)
                self._err_buffer["CAN_ErrorFrame.ESI"] = int(msg.error_state_indicator)
                self._err_buffer["CAN_ErrorFrame.BRS"] = int(msg.bitrate_switch)
                self._err_buffer["CAN_ErrorFrame.EDL"] = 1
            else:
                self._err_buffer["CAN_ErrorFrame.DLC"] = msg.dlc
                self._err_buffer["CAN_ErrorFrame.ESI"] = 0
                self._err_buffer["CAN_ErrorFrame.BRS"] = 0
                self._err_buffer["CAN_ErrorFrame.EDL"] = 0

            sigs = [(np.array([timestamp]), None), (self._err_buffer, None)]
            self._mdf.extend(1, sigs)

        else:
            if channel is not None:
                self._std_buffer["CAN_DataFrame.BusChannel"] = channel

            self._std_buffer["CAN_DataFrame.ID"] = msg.arbitration_id
            self._std_buffer["CAN_DataFrame.IDE"] = int(msg.is_extended_id)
            self._std_buffer["CAN_DataFrame.Dir"] = 0 if msg.is_rx else 1
            data = msg.data
            size = len(data)
            self._std_buffer["CAN_DataFrame.DataLength"] = size
            self._std_buffer["CAN_DataFrame.DataBytes"][0, :size] = data
            if msg.is_fd:
                self._std_buffer["CAN_DataFrame.DLC"] = len2dlc(msg.dlc)
                self._std_buffer["CAN_DataFrame.ESI"] = int(msg.error_state_indicator)
                self._std_buffer["CAN_DataFrame.BRS"] = int(msg.bitrate_switch)
                self._std_buffer["CAN_DataFrame.EDL"] = 1
            else:
                self._std_buffer["CAN_DataFrame.DLC"] = msg.dlc
                self._std_buffer["CAN_DataFrame.ESI"] = 0
                self._std_buffer["CAN_DataFrame.BRS"] = 0
                self._std_buffer["CAN_DataFrame.EDL"] = 0

            sigs = [(np.array([timestamp]), None), (self._std_buffer, None)]
            self._mdf.extend(0, sigs)

        # reset buffer structure
        self._std_buffer = np.zeros(1, dtype=STD_DTYPE)
        self._err_buffer = np.zeros(1, dtype=ERR_DTYPE)
        self._rtr_buffer = np.zeros(1, dtype=RTR_DTYPE)


class FrameIterator(abc.ABC):
    """
    Iterator helper class for common handling among CAN DataFrames, ErrorFrames and RemoteFrames.
    """

    # Number of records to request for each asammdf call
    _chunk_size = 1000

    def __init__(self, mdf: MDF4, group_index: int, start_timestamp: float, name: str):
        self._mdf = mdf
        self._group_index = group_index
        self._start_timestamp = start_timestamp
        self._name = name

        # Extract names
        channel_group: ChannelGroup = self._mdf.groups[self._group_index]

        self._channel_names = []

        for channel in channel_group.channels:
            if str(channel.name).startswith(f"{self._name}."):
                self._channel_names.append(channel.name)

    def _get_data(self, current_offset: int) -> Signal:
        # NOTE: asammdf suggests using select instead of get. Select seem to miss converting some
        #       channels which get does convert as expected.
        data_raw = self._mdf.get(
            self._name,
            self._group_index,
            record_offset=current_offset,
            record_count=self._chunk_size,
            raw=False,
        )

        return data_raw

    @abc.abstractmethod
    def __iter__(self) -> Generator[Message, None, None]:
        pass


class MF4Reader(BinaryIOMessageReader):
    """
    Iterator of CAN messages from a MF4 logging file.

    The MF4Reader only supports MF4 files with CAN bus logging.
    """

    # NOTE: Readout based on the bus logging code from asammdf GUI

    class _CANDataFrameIterator(FrameIterator):

        def __init__(self, mdf: MDF4, group_index: int, start_timestamp: float):
            super().__init__(mdf, group_index, start_timestamp, "CAN_DataFrame")

        def __iter__(self) -> Generator[Message, None, None]:
            for current_offset in range(
                0,
                self._mdf.groups[self._group_index].channel_group.cycles_nr,
                self._chunk_size,
            ):
                data = self._get_data(current_offset)
                names = data.samples[0].dtype.names

                for i in range(len(data)):
                    data_length = int(data["CAN_DataFrame.DataLength"][i])

                    kv: dict[str, Any] = {
                        "timestamp": float(data.timestamps[i]) + self._start_timestamp,
                        "arbitration_id": int(data["CAN_DataFrame.ID"][i]) & 0x1FFFFFFF,
                        "data": data["CAN_DataFrame.DataBytes"][i][
                            :data_length
                        ].tobytes(),
                    }

                    if "CAN_DataFrame.BusChannel" in names:
                        kv["channel"] = int(data["CAN_DataFrame.BusChannel"][i])
                    if "CAN_DataFrame.Dir" in names:
                        if data["CAN_DataFrame.Dir"][i].dtype.kind == "S":
                            kv["is_rx"] = data["CAN_DataFrame.Dir"][i] == b"Rx"
                        else:
                            kv["is_rx"] = int(data["CAN_DataFrame.Dir"][i]) == 0
                    if "CAN_DataFrame.IDE" in names:
                        kv["is_extended_id"] = bool(data["CAN_DataFrame.IDE"][i])
                    if "CAN_DataFrame.EDL" in names:
                        kv["is_fd"] = bool(data["CAN_DataFrame.EDL"][i])
                    if "CAN_DataFrame.BRS" in names:
                        kv["bitrate_switch"] = bool(data["CAN_DataFrame.BRS"][i])
                    if "CAN_DataFrame.ESI" in names:
                        kv["error_state_indicator"] = bool(data["CAN_DataFrame.ESI"][i])

                    yield Message(**kv)

    class _CANErrorFrameIterator(FrameIterator):

        def __init__(self, mdf: MDF4, group_index: int, start_timestamp: float):
            super().__init__(mdf, group_index, start_timestamp, "CAN_ErrorFrame")

        def __iter__(self) -> Generator[Message, None, None]:
            for current_offset in range(
                0,
                self._mdf.groups[self._group_index].channel_group.cycles_nr,
                self._chunk_size,
            ):
                data = self._get_data(current_offset)
                names = data.samples[0].dtype.names

                for i in range(len(data)):
                    kv: dict[str, Any] = {
                        "timestamp": float(data.timestamps[i]) + self._start_timestamp,
                        "is_error_frame": True,
                    }

                    if "CAN_ErrorFrame.BusChannel" in names:
                        kv["channel"] = int(data["CAN_ErrorFrame.BusChannel"][i])
                    if "CAN_ErrorFrame.Dir" in names:
                        if data["CAN_ErrorFrame.Dir"][i].dtype.kind == "S":
                            kv["is_rx"] = data["CAN_ErrorFrame.Dir"][i] == b"Rx"
                        else:
                            kv["is_rx"] = int(data["CAN_ErrorFrame.Dir"][i]) == 0
                    if "CAN_ErrorFrame.ID" in names:
                        kv["arbitration_id"] = (
                            int(data["CAN_ErrorFrame.ID"][i]) & 0x1FFFFFFF
                        )
                    if "CAN_ErrorFrame.IDE" in names:
                        kv["is_extended_id"] = bool(data["CAN_ErrorFrame.IDE"][i])
                    if "CAN_ErrorFrame.EDL" in names:
                        kv["is_fd"] = bool(data["CAN_ErrorFrame.EDL"][i])
                    if "CAN_ErrorFrame.BRS" in names:
                        kv["bitrate_switch"] = bool(data["CAN_ErrorFrame.BRS"][i])
                    if "CAN_ErrorFrame.ESI" in names:
                        kv["error_state_indicator"] = bool(
                            data["CAN_ErrorFrame.ESI"][i]
                        )
                    if "CAN_ErrorFrame.RTR" in names:
                        kv["is_remote_frame"] = bool(data["CAN_ErrorFrame.RTR"][i])
                    if (
                        "CAN_ErrorFrame.DataLength" in names
                        and "CAN_ErrorFrame.DataBytes" in names
                    ):
                        data_length = int(data["CAN_ErrorFrame.DataLength"][i])
                        kv["data"] = data["CAN_ErrorFrame.DataBytes"][i][
                            :data_length
                        ].tobytes()

                    yield Message(**kv)

    class _CANRemoteFrameIterator(FrameIterator):

        def __init__(self, mdf: MDF4, group_index: int, start_timestamp: float):
            super().__init__(mdf, group_index, start_timestamp, "CAN_RemoteFrame")

        def __iter__(self) -> Generator[Message, None, None]:
            for current_offset in range(
                0,
                self._mdf.groups[self._group_index].channel_group.cycles_nr,
                self._chunk_size,
            ):
                data = self._get_data(current_offset)
                names = data.samples[0].dtype.names

                for i in range(len(data)):
                    kv: dict[str, Any] = {
                        "timestamp": float(data.timestamps[i]) + self._start_timestamp,
                        "arbitration_id": int(data["CAN_RemoteFrame.ID"][i])
                        & 0x1FFFFFFF,
                        "dlc": int(data["CAN_RemoteFrame.DLC"][i]),
                        "is_remote_frame": True,
                    }

                    if "CAN_RemoteFrame.BusChannel" in names:
                        kv["channel"] = int(data["CAN_RemoteFrame.BusChannel"][i])
                    if "CAN_RemoteFrame.Dir" in names:
                        if data["CAN_RemoteFrame.Dir"][i].dtype.kind == "S":
                            kv["is_rx"] = data["CAN_RemoteFrame.Dir"][i] == b"Rx"
                        else:
                            kv["is_rx"] = int(data["CAN_RemoteFrame.Dir"][i]) == 0
                    if "CAN_RemoteFrame.IDE" in names:
                        kv["is_extended_id"] = bool(data["CAN_RemoteFrame.IDE"][i])

                    yield Message(**kv)

    def __init__(
        self,
        file: Union[StringPathLike, BinaryIO],
        **kwargs: Any,
    ) -> None:
        """
        :param file: a path-like object or as file-like object to read from
                        If this is a file-like object, is has to be opened in
                        binary read mode, not text read mode.
        """
        if asammdf is None:
            raise NotImplementedError(
                "The asammdf package was not found. Install python-can with "
                "the optional dependency [mf4] to use the MF4Reader."
            )

        super().__init__(file, mode="rb")

        self._mdf: MDF4
        if isinstance(file, BufferedIOBase):
            self._mdf = cast("MDF4", MDF(BytesIO(file.read())))
        else:
            self._mdf = cast("MDF4", MDF(file))

        self._start_timestamp = self._mdf.header.start_time.timestamp()

    def __iter__(self) -> Iterator[Message]:
        # To handle messages split over multiple channel groups, create a single iterator per
        # channel group and merge these iterators into a single iterator using heapq.
        iterators: list[FrameIterator] = []
        for group_index, group in enumerate(self._mdf.groups):
            channel_group: ChannelGroup = group.channel_group

            if not channel_group.flags & FLAG_CG_BUS_EVENT:
                # Not a bus event, skip
                continue

            if channel_group.cycles_nr == 0:
                # No data, skip
                continue

            acquisition_source: Optional[Source] = channel_group.acq_source

            if acquisition_source is None:
                # No source information, skip
                continue
            if not acquisition_source.source_type & Source.SOURCE_BUS:
                # Not a bus type (likely already covered by the channel group flag), skip
                continue

            channel_names = [channel.name for channel in group.channels]

            if acquisition_source.bus_type == Source.BUS_TYPE_CAN:
                if "CAN_DataFrame" in channel_names:
                    iterators.append(
                        self._CANDataFrameIterator(
                            self._mdf, group_index, self._start_timestamp
                        )
                    )
                elif "CAN_ErrorFrame" in channel_names:
                    iterators.append(
                        self._CANErrorFrameIterator(
                            self._mdf, group_index, self._start_timestamp
                        )
                    )
                elif "CAN_RemoteFrame" in channel_names:
                    iterators.append(
                        self._CANRemoteFrameIterator(
                            self._mdf, group_index, self._start_timestamp
                        )
                    )
            else:
                # Unknown bus type, skip
                continue

        # Create merged iterator over all the groups, using the timestamps as comparison key
        return iter(heapq.merge(*iterators, key=lambda x: x.timestamp))

    def stop(self) -> None:
        self._mdf.close()
        self._mdf = None
        super().stop()



""" """

__all__ = [
    "Usb2CanAbstractionLayer",
    "Usb2canBus",
    "serial_selector",
    "usb2canInterface",
    "usb2canabstractionlayer",
]

from .usb2canabstractionlayer import Usb2CanAbstractionLayer
from .usb2canInterface import Usb2canBus



from contextlib import nullcontext
from threading import RLock
from typing import Any, Optional

from can import typechecking
from can.bus import BusABC, BusState, CanProtocol
from can.message import Message

from .interface import Bus

try:
    # Only raise an exception on instantiation but allow module
    # to be imported
    from wrapt import ObjectProxy

    import_exc = None
except ImportError as exc:
    ObjectProxy = object
    import_exc = exc


class ThreadSafeBus(ObjectProxy):  # pylint: disable=abstract-method
    """
    Contains a thread safe :class:`can.BusABC` implementation that
    wraps around an existing interface instance. All public methods
    of that base class are now safe to be called from multiple threads.
    The send and receive methods are synchronized separately.

    Use this as a drop-in replacement for :class:`~can.BusABC`.

    .. note::

        This approach assumes that both :meth:`~can.BusABC.send` and
        :meth:`~can.BusABC._recv_internal` of the underlying bus instance can be
        called simultaneously, and that the methods use :meth:`~can.BusABC._recv_internal`
        instead of :meth:`~can.BusABC.recv` directly.
    """

    __wrapped__: BusABC

    def __init__(
        self,
        channel: Optional[typechecking.Channel] = None,
        interface: Optional[str] = None,
        config_context: Optional[str] = None,
        ignore_config: bool = False,
        **kwargs: Any,
    ) -> None:
        if import_exc is not None:
            raise import_exc

        super().__init__(
            Bus(
                channel=channel,
                interface=interface,
                config_context=config_context,
                ignore_config=ignore_config,
                **kwargs,
            )
        )

        # now, BusABC.send_periodic() does not need a lock anymore, but the
        # implementation still requires a context manager
        self.__wrapped__._lock_send_periodic = nullcontext()  # type: ignore[assignment]

        # init locks for sending and receiving separately
        self._lock_send = RLock()
        self._lock_recv = RLock()

    def recv(self, timeout: Optional[float] = None) -> Optional[Message]:
        with self._lock_recv:
            return self.__wrapped__.recv(timeout=timeout)

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        with self._lock_send:
            return self.__wrapped__.send(msg=msg, timeout=timeout)

    # send_periodic does not need a lock, since the underlying
    # `send` method is already synchronized

    @property
    def filters(self) -> Optional[typechecking.CanFilters]:
        with self._lock_recv:
            return self.__wrapped__.filters

    @filters.setter
    def filters(self, filters: Optional[typechecking.CanFilters]) -> None:
        with self._lock_recv:
            self.__wrapped__.filters = filters

    def set_filters(self, filters: Optional[typechecking.CanFilters] = None) -> None:
        with self._lock_recv:
            return self.__wrapped__.set_filters(filters=filters)

    def flush_tx_buffer(self) -> None:
        with self._lock_send:
            return self.__wrapped__.flush_tx_buffer()

    def shutdown(self) -> None:
        with self._lock_send, self._lock_recv:
            return self.__wrapped__.shutdown()

    @property
    def state(self) -> BusState:
        with self._lock_send, self._lock_recv:
            return self.__wrapped__.state

    @state.setter
    def state(self, new_state: BusState) -> None:
        with self._lock_send, self._lock_recv:
            self.__wrapped__.state = new_state

    @property
    def protocol(self) -> CanProtocol:
        with self._lock_send, self._lock_recv:
            return self.__wrapped__.protocol



import argparse
import re
from collections.abc import Sequence
from typing import Any, Optional, Union

import can
from can.typechecking import CanFilter, TAdditionalCliArgs
from can.util import _dict2timing, cast_from_string


def add_bus_arguments(
    parser: argparse.ArgumentParser,
    *,
    filter_arg: bool = False,
    prefix: Optional[str] = None,
    group_title: Optional[str] = None,
) -> None:
    """Adds CAN bus configuration options to an argument parser.

    :param parser:
        The argument parser to which the options will be added.
    :param filter_arg:
        Whether to include the filter argument.
    :param prefix:
        An optional prefix for the argument names, allowing configuration of multiple buses.
    :param group_title:
        The title of the argument group. If not provided, a default title will be generated
        based on the prefix. For example, "bus arguments (prefix)" if a prefix is specified,
        or "bus arguments" otherwise.
    """
    if group_title is None:
        group_title = f"bus arguments ({prefix})" if prefix else "bus arguments"

    group = parser.add_argument_group(group_title)

    flags = [f"--{prefix}-channel"] if prefix else ["-c", "--channel"]
    dest = f"{prefix}_channel" if prefix else "channel"
    group.add_argument(
        *flags,
        dest=dest,
        default=argparse.SUPPRESS,
        metavar="CHANNEL",
        help=r"Most backend interfaces require some sort of channel. For "
        r"example with the serial interface the channel might be a rfcomm"
        r' device: "/dev/rfcomm0". With the socketcan interface valid '
        r'channel examples include: "can0", "vcan0".',
    )

    flags = [f"--{prefix}-interface"] if prefix else ["-i", "--interface"]
    dest = f"{prefix}_interface" if prefix else "interface"
    group.add_argument(
        *flags,
        dest=dest,
        default=argparse.SUPPRESS,
        choices=sorted(can.VALID_INTERFACES),
        help="""Specify the backend CAN interface to use. If left blank,
                        fall back to reading from configuration files.""",
    )

    flags = [f"--{prefix}-bitrate"] if prefix else ["-b", "--bitrate"]
    dest = f"{prefix}_bitrate" if prefix else "bitrate"
    group.add_argument(
        *flags,
        dest=dest,
        type=int,
        default=argparse.SUPPRESS,
        metavar="BITRATE",
        help="Bitrate to use for the CAN bus.",
    )

    flags = [f"--{prefix}-fd"] if prefix else ["--fd"]
    dest = f"{prefix}_fd" if prefix else "fd"
    group.add_argument(
        *flags,
        dest=dest,
        default=argparse.SUPPRESS,
        action="store_true",
        help="Activate CAN-FD support",
    )

    flags = [f"--{prefix}-data-bitrate"] if prefix else ["--data-bitrate"]
    dest = f"{prefix}_data_bitrate" if prefix else "data_bitrate"
    group.add_argument(
        *flags,
        dest=dest,
        type=int,
        default=argparse.SUPPRESS,
        metavar="DATA_BITRATE",
        help="Bitrate to use for the data phase in case of CAN-FD.",
    )

    flags = [f"--{prefix}-timing"] if prefix else ["--timing"]
    dest = f"{prefix}_timing" if prefix else "timing"
    group.add_argument(
        *flags,
        dest=dest,
        action=_BitTimingAction,
        nargs=argparse.ONE_OR_MORE,
        default=argparse.SUPPRESS,
        metavar="TIMING_ARG",
        help="Configure bit rate and bit timing. For example, use "
        "`--timing f_clock=8_000_000 tseg1=5 tseg2=2 sjw=2 brp=2 nof_samples=1` for classical CAN "
        "or `--timing f_clock=80_000_000 nom_tseg1=119 nom_tseg2=40 nom_sjw=40 nom_brp=1 "
        "data_tseg1=29 data_tseg2=10 data_sjw=10 data_brp=1` for CAN FD. "
        "Check the python-can documentation to verify whether your "
        "CAN interface supports the `timing` argument.",
    )

    if filter_arg:
        flags = [f"--{prefix}-filter"] if prefix else ["--filter"]
        dest = f"{prefix}_can_filters" if prefix else "can_filters"
        group.add_argument(
            *flags,
            dest=dest,
            nargs=argparse.ONE_OR_MORE,
            action=_CanFilterAction,
            default=argparse.SUPPRESS,
            metavar="{<can_id>:<can_mask>,<can_id>~<can_mask>}",
            help="R|Space separated CAN filters for the given CAN interface:"
            "\n      <can_id>:<can_mask> (matches when <received_can_id> & mask =="
            " can_id & mask)"
            "\n      <can_id>~<can_mask> (matches when <received_can_id> & mask !="
            " can_id & mask)"
            "\nFx to show only frames with ID 0x100 to 0x103 and 0x200 to 0x20F:"
            "\n      python -m can.viewer --filter 100:7FC 200:7F0"
            "\nNote that the ID and mask are always interpreted as hex values",
        )

    flags = [f"--{prefix}-bus-kwargs"] if prefix else ["--bus-kwargs"]
    dest = f"{prefix}_bus_kwargs" if prefix else "bus_kwargs"
    group.add_argument(
        *flags,
        dest=dest,
        action=_BusKwargsAction,
        nargs=argparse.ONE_OR_MORE,
        default=argparse.SUPPRESS,
        metavar="BUS_KWARG",
        help="Pass keyword arguments down to the instantiation of the bus class. "
        "For example, `-i vector -c 1 --bus-kwargs app_name=MyCanApp serial=1234` is equivalent "
        "to opening the bus with `can.Bus('vector', channel=1, app_name='MyCanApp', serial=1234)",
    )


def create_bus_from_namespace(
    namespace: argparse.Namespace,
    *,
    prefix: Optional[str] = None,
    **kwargs: Any,
) -> can.BusABC:
    """Creates and returns a CAN bus instance based on the provided namespace and arguments.

    :param namespace:
        The namespace containing parsed arguments.
    :param prefix:
        An optional prefix for the argument names, enabling support for multiple buses.
    :param kwargs:
        Additional keyword arguments to configure the bus.
    :return:
        A CAN bus instance.
    """
    config: dict[str, Any] = {"single_handle": True, **kwargs}

    for keyword in (
        "channel",
        "interface",
        "bitrate",
        "fd",
        "data_bitrate",
        "can_filters",
        "timing",
        "bus_kwargs",
    ):
        prefixed_keyword = f"{prefix}_{keyword}" if prefix else keyword

        if prefixed_keyword in namespace:
            value = getattr(namespace, prefixed_keyword)

            if keyword == "bus_kwargs":
                config.update(value)
            else:
                config[keyword] = value

    try:
        return can.Bus(**config)
    except Exception as exc:
        err_msg = f"Unable to instantiate bus from arguments {vars(namespace)}."
        raise argparse.ArgumentError(None, err_msg) from exc


class _CanFilterAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, list):
            raise argparse.ArgumentError(self, "Invalid filter argument")

        print(f"Adding filter(s): {values}")
        can_filters: list[CanFilter] = []

        for filt in values:
            if ":" in filt:
                parts = filt.split(":")
                can_id = int(parts[0], base=16)
                can_mask = int(parts[1], base=16)
            elif "~" in filt:
                parts = filt.split("~")
                can_id = int(parts[0], base=16) | 0x20000000  # CAN_INV_FILTER
                can_mask = int(parts[1], base=16) & 0x20000000  # socket.CAN_ERR_FLAG
            else:
                raise argparse.ArgumentError(self, "Invalid filter argument")
            can_filters.append({"can_id": can_id, "can_mask": can_mask})

        setattr(namespace, self.dest, can_filters)


class _BitTimingAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, list):
            raise argparse.ArgumentError(self, "Invalid --timing argument")

        timing_dict: dict[str, int] = {}
        for arg in values:
            try:
                key, value_string = arg.split("=")
                value = int(value_string)
                timing_dict[key] = value
            except ValueError:
                raise argparse.ArgumentError(
                    self, f"Invalid timing argument: {arg}"
                ) from None

        if not (timing := _dict2timing(timing_dict)):
            err_msg = "Invalid --timing argument. Incomplete parameters."
            raise argparse.ArgumentError(self, err_msg)

        setattr(namespace, self.dest, timing)
        print(timing)


class _BusKwargsAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        if not isinstance(values, list):
            raise argparse.ArgumentError(self, "Invalid --bus-kwargs argument")

        bus_kwargs: dict[str, Union[str, int, float, bool]] = {}

        for arg in values:
            try:
                match = re.match(
                    r"^(?P<name>[_a-zA-Z][_a-zA-Z0-9]*)=(?P<value>\S*?)$",
                    arg,
                )
                if not match:
                    raise ValueError
                key = match["name"].replace("-", "_")
                string_val = match["value"]
                bus_kwargs[key] = cast_from_string(string_val)
            except ValueError:
                raise argparse.ArgumentError(
                    self,
                    f"Unable to parse bus keyword argument '{arg}'",
                ) from None

        setattr(namespace, self.dest, bus_kwargs)


def _add_extra_args(
    parser: Union[argparse.ArgumentParser, argparse._ArgumentGroup],
) -> None:
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="The remaining arguments will be used for logger/player initialisation. "
        "For example, `can_logger -i virtual -c test -f logfile.blf --compression-level=9` "
        "passes the keyword argument `compression_level=9` to the BlfWriter.",
    )


def _parse_additional_config(unknown_args: Sequence[str]) -> TAdditionalCliArgs:
    for arg in unknown_args:
        if not re.match(r"^--[a-zA-Z][a-zA-Z0-9\-]*=\S*?$", arg):
            raise ValueError(f"Parsing argument {arg} failed")

    def _split_arg(_arg: str) -> tuple[str, str]:
        left, right = _arg.split("=", 1)
        return left.lstrip("-").replace("-", "_"), right

    args: dict[str, Union[str, int, float, bool]] = {}
    for key, string_val in map(_split_arg, unknown_args):
        args[key] = cast_from_string(string_val)
    return args


def _set_logging_level_from_namespace(namespace: argparse.Namespace) -> None:
    if "verbosity" in namespace:
        logging_level_names = [
            "critical",
            "error",
            "warning",
            "info",
            "debug",
            "subdebug",
        ]
        can.set_logging_level(logging_level_names[min(5, namespace.verbosity)])



import importlib
import sys
from dataclasses import dataclass
from importlib.metadata import entry_points
from typing import Any


@dataclass
class _EntryPoint:
    key: str
    module_name: str
    class_name: str

    def load(self) -> Any:
        module = importlib.import_module(self.module_name)
        return getattr(module, self.class_name)


# See https://docs.python.org/3/library/importlib.metadata.html#entry-points,
# "Compatibility Note".
if sys.version_info >= (3, 10):

    def read_entry_points(group: str) -> list[_EntryPoint]:
        return [
            _EntryPoint(ep.name, ep.module, ep.attr) for ep in entry_points(group=group)
        ]

else:

    def read_entry_points(group: str) -> list[_EntryPoint]:
        return [
            _EntryPoint(ep.name, *ep.value.split(":", maxsplit=1))
            for ep in entry_points().get(group, [])  # pylint: disable=no-member
        ]



"""
See the :class:`Logger` class.
"""

import gzip
import os
import pathlib
from abc import ABC, abstractmethod
from datetime import datetime
from types import TracebackType
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
    Literal,
    Optional,
)

from typing_extensions import Self

from .._entry_points import read_entry_points
from ..message import Message
from ..typechecking import StringPathLike
from .asc import ASCWriter
from .blf import BLFWriter
from .canutils import CanutilsLogWriter
from .csv import CSVWriter
from .generic import (
    BinaryIOMessageWriter,
    FileIOMessageWriter,
    MessageWriter,
    SizedMessageWriter,
    TextIOMessageWriter,
)
from .mf4 import MF4Writer
from .printer import Printer
from .sqlite import SqliteWriter
from .trc import TRCWriter

#: A map of file suffixes to their corresponding
#: :class:`can.io.generic.MessageWriter` class
MESSAGE_WRITERS: Final[dict[str, type[MessageWriter]]] = {
    ".asc": ASCWriter,
    ".blf": BLFWriter,
    ".csv": CSVWriter,
    ".db": SqliteWriter,
    ".log": CanutilsLogWriter,
    ".mf4": MF4Writer,
    ".trc": TRCWriter,
    ".txt": Printer,
}


def _update_writer_plugins() -> None:
    """Update available message writer plugins from entry points."""
    for entry_point in read_entry_points("can.io.message_writer"):
        if entry_point.key in MESSAGE_WRITERS:
            continue

        writer_class = entry_point.load()
        if issubclass(writer_class, MessageWriter):
            MESSAGE_WRITERS[entry_point.key] = writer_class


def _get_logger_for_suffix(suffix: str) -> type[MessageWriter]:
    try:
        return MESSAGE_WRITERS[suffix]
    except KeyError:
        raise ValueError(
            f'No write support for unknown log format "{suffix}"'
        ) from None


def _compress(filename: StringPathLike, **kwargs: Any) -> FileIOMessageWriter[Any]:
    """
    Return the suffix and io object of the decompressed file.
    File will automatically recompress upon close.
    """
    suffixes = pathlib.Path(filename).suffixes
    if len(suffixes) != 2:
        raise ValueError(
            f"No write support for unknown log format \"{''.join(suffixes)}\""
        ) from None

    real_suffix = suffixes[-2].lower()
    if real_suffix in (".blf", ".db"):
        raise ValueError(
            f"The file type {real_suffix} is currently incompatible with gzip."
        )
    logger_type = _get_logger_for_suffix(real_suffix)
    append = kwargs.get("append", False)

    if issubclass(logger_type, BinaryIOMessageWriter):
        return logger_type(
            file=gzip.open(filename=filename, mode="ab" if append else "wb"), **kwargs
        )

    elif issubclass(logger_type, TextIOMessageWriter):
        return logger_type(
            file=gzip.open(filename=filename, mode="at" if append else "wt"), **kwargs
        )

    raise ValueError(
        f"The file type {real_suffix} is currently incompatible with gzip."
    )


def Logger(  # noqa: N802
    filename: Optional[StringPathLike], **kwargs: Any
) -> MessageWriter:
    """Find and return the appropriate :class:`~can.io.generic.MessageWriter` instance
    for a given file suffix.

    The format is determined from the file suffix which can be one of:
      * .asc :class:`can.ASCWriter`
      * .blf :class:`can.BLFWriter`
      * .csv: :class:`can.CSVWriter`
      * .db :class:`can.SqliteWriter`
      * .log :class:`can.CanutilsLogWriter`
      * .mf4 :class:`can.MF4Writer`
        (optional, depends on `asammdf <https://github.com/danielhrisca/asammdf>`_)
      * .trc :class:`can.TRCWriter`
      * .txt :class:`can.Printer`

    Any of these formats can be used with gzip compression by appending
    the suffix .gz (e.g. filename.asc.gz). However, third-party tools might not
    be able to read these files.

    The **filename** may also be *None*, to fall back to :class:`can.Printer`.

    The log files may be incomplete until `stop()` is called due to buffering.

    :param filename:
        the filename/path of the file to write to,
        may be a path-like object or None to
        instantiate a :class:`~can.Printer`
    :raises ValueError:
        if the filename's suffix is of an unknown file type

    .. note::
        This function itself is just a dispatcher, and any positional and keyword
        arguments are passed on to the returned instance.
    """

    if filename is None:
        return Printer(**kwargs)

    _update_writer_plugins()

    suffix = pathlib.PurePath(filename).suffix.lower()
    if suffix == ".gz":
        return _compress(filename, **kwargs)

    logger_type = _get_logger_for_suffix(suffix)
    return logger_type(file=filename, **kwargs)


class BaseRotatingLogger(MessageWriter, ABC):
    """
    Base class for rotating CAN loggers. This class is not meant to be
    instantiated directly. Subclasses must implement the :meth:`should_rollover`
    and :meth:`do_rollover` methods according to their rotation strategy.

    The rotation behavior can be further customized by the user by setting
    the :attr:`namer` and :attr:`rotator` attributes after instantiating the subclass.

    These attributes as well as the methods :meth:`rotation_filename` and :meth:`rotate`
    and the corresponding docstrings are carried over from the python builtin
    :class:`~logging.handlers.BaseRotatingHandler`.

    Subclasses must set the `_writer` attribute upon initialization.
    """

    _supported_formats: ClassVar[set[str]] = set()

    #: If this attribute is set to a callable, the :meth:`~BaseRotatingLogger.rotation_filename`
    #: method delegates to this callable. The parameters passed to the callable are
    #: those passed to :meth:`~BaseRotatingLogger.rotation_filename`.
    namer: Optional[Callable[[StringPathLike], StringPathLike]] = None

    #: If this attribute is set to a callable, the :meth:`~BaseRotatingLogger.rotate` method
    #: delegates to this callable. The parameters passed to the callable are those
    #: passed to :meth:`~BaseRotatingLogger.rotate`.
    rotator: Optional[Callable[[StringPathLike, StringPathLike], None]] = None

    #: An integer counter to track the number of rollovers.
    rollover_count: int = 0

    def __init__(self, **kwargs: Any) -> None:
        self.writer_kwargs = kwargs

    @property
    @abstractmethod
    def writer(self) -> MessageWriter:
        """This attribute holds an instance of a writer class which manages the actual file IO."""
        raise NotImplementedError

    def rotation_filename(self, default_name: StringPathLike) -> StringPathLike:
        """Modify the filename of a log file when rotating.

        This is provided so that a custom filename can be provided.
        The default implementation calls the :attr:`namer` attribute of the
        handler, if it's callable, passing the default name to
        it. If the attribute isn't callable (the default is :obj:`None`), the name
        is returned unchanged.

        :param default_name:
            The default name for the log file.
        """
        if not callable(self.namer):
            return default_name

        return self.namer(default_name)  # pylint: disable=not-callable

    def rotate(self, source: StringPathLike, dest: StringPathLike) -> None:
        """When rotating, rotate the current log.

        The default implementation calls the :attr:`rotator` attribute of the
        handler, if it's callable, passing the `source` and `dest` arguments to
        it. If the attribute isn't callable (the default is :obj:`None`), the source
        is simply renamed to the destination.

        :param source:
            The source filename. This is normally the base
            filename, e.g. `"test.log"`
        :param dest:
            The destination filename. This is normally
            what the source is rotated to, e.g. `"test_#001.log"`.
        """
        if not callable(self.rotator):
            if os.path.exists(source):
                os.rename(source, dest)
        else:
            self.rotator(source, dest)  # pylint: disable=not-callable

    def on_message_received(self, msg: Message) -> None:
        """This method is called to handle the given message.

        :param msg:
            the delivered message
        """
        if self.should_rollover(msg):
            self.do_rollover()
            self.rollover_count += 1

        self.writer.on_message_received(msg)

    def _get_new_writer(self, filename: StringPathLike) -> MessageWriter:
        """Instantiate a new writer.

        .. note::
            The :attr:`self.writer` should be closed prior to calling this function.

        :param filename:
            Path-like object that specifies the location and name of the log file.
            The log file format is defined by the suffix of `filename`.
        :return:
            An instance of a writer class.
        """
        suffixes = pathlib.Path(filename).suffixes
        for suffix_length in range(len(suffixes), 0, -1):
            suffix = "".join(suffixes[-suffix_length:]).lower()
            if suffix not in self._supported_formats:
                continue
            logger = Logger(filename=filename, **self.writer_kwargs)
            return logger

        raise ValueError(
            f'The log format of "{pathlib.Path(filename).name}" '
            f"is not supported by {self.__class__.__name__}. "
            f"{self.__class__.__name__} supports the following formats: "
            f"{', '.join(self._supported_formats)}"
        )

    def stop(self) -> None:
        """Stop handling new messages.

        Carry out any final tasks to ensure
        data is persisted and cleanup any open resources.
        """
        self.writer.stop()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        self.stop()
        return False

    @abstractmethod
    def should_rollover(self, msg: Message) -> bool:
        """Determine if the rollover conditions are met."""

    @abstractmethod
    def do_rollover(self) -> None:
        """Perform rollover."""


class SizedRotatingLogger(BaseRotatingLogger):
    """Log CAN messages to a sequence of files with a given maximum size.

    The logger creates a log file with the given `base_filename`. When the
    size threshold is reached the current log file is closed and renamed
    by adding a timestamp and the rollover count. A new log file is then
    created and written to.

    This behavior can be customized by setting the
    :attr:`~can.io.BaseRotatingLogger.namer` and
    :attr:`~can.io.BaseRotatingLogger.rotator`
    attribute.

    Example::

        from can import Notifier, SizedRotatingLogger
        from can.interfaces.vector import VectorBus

        bus = VectorBus(channel=[0], app_name="CANape", fd=True)

        logger = SizedRotatingLogger(
            base_filename="my_logfile.asc",
            max_bytes=5 * 1024 ** 2,  # =5MB
        )
        logger.rollover_count = 23  # start counter at 23

        notifier = Notifier(bus=bus, listeners=[logger])

    The SizedRotatingLogger currently supports the formats
      * .asc: :class:`can.ASCWriter`
      * .blf :class:`can.BLFWriter`
      * .csv: :class:`can.CSVWriter`
      * .log :class:`can.CanutilsLogWriter`
      * .txt :class:`can.Printer` (if pointing to a file)

    .. note::
        The :class:`can.SqliteWriter` is not supported yet.

    The log files on disk may be incomplete due to buffering until
    :meth:`~can.Listener.stop` is called.
    """

    _supported_formats: ClassVar[set[str]] = {".asc", ".blf", ".csv", ".log", ".txt"}

    def __init__(
        self,
        base_filename: StringPathLike,
        max_bytes: int = 0,
        **kwargs: Any,
    ) -> None:
        """
        :param base_filename:
            A path-like object for the base filename. The log file format is defined by
            the suffix of `base_filename`.
        :param max_bytes:
            The size threshold at which a new log file shall be created. If set to 0, no
            rollover will be performed.
        """
        super().__init__(**kwargs)

        self.base_filename = os.path.abspath(base_filename)
        self.max_bytes = max_bytes

        self._writer = self._get_new_writer(self.base_filename)

    def _get_new_writer(self, filename: StringPathLike) -> SizedMessageWriter:
        writer = super()._get_new_writer(filename)
        if isinstance(writer, SizedMessageWriter):
            return writer
        raise TypeError

    @property
    def writer(self) -> SizedMessageWriter:
        return self._writer

    def should_rollover(self, msg: Message) -> bool:
        if self.max_bytes <= 0:
            return False

        file_size = self.writer.file_size()
        if file_size is None:
            return False

        return file_size >= self.max_bytes

    def do_rollover(self) -> None:
        if self.writer:
            self.writer.stop()

        sfn = self.base_filename
        dfn = self.rotation_filename(self._default_name())
        self.rotate(sfn, dfn)

        self._writer = self._get_new_writer(self.base_filename)

    def _default_name(self) -> StringPathLike:
        """Generate the default rotation filename."""
        path = pathlib.Path(self.base_filename)
        new_name = (
            path.stem.split(".")[0]
            + "_"
            + datetime.now().strftime("%Y-%m-%dT%H%M%S")
            + "_"
            + f"#{self.rollover_count:03}"
            + "".join(path.suffixes[-2:])
        )
        return str(path.parent / new_name)



"""
A text based interface. For example use over serial ports like
"/dev/ttyS1" or "/dev/ttyUSB0" on Linux machines or "COM1" on Windows.
The interface is a simple implementation that has been used for
recording CAN traces.

See the interface documentation for the format being used.
"""

import io
import logging
import struct
from collections.abc import Sequence
from typing import Any, Optional, cast

from can import (
    BusABC,
    CanInitializationError,
    CanInterfaceNotImplementedError,
    CanOperationError,
    CanProtocol,
    CanTimeoutError,
    Message,
)
from can.typechecking import AutoDetectedConfig

logger = logging.getLogger("can.serial")

try:
    import serial.tools.list_ports
except ImportError:
    logger.warning(
        "You won't be able to use the serial can backend without "
        "the `pyserial` package installed!"
    )
    serial = None


CAN_ERR_FLAG = 0x20000000
CAN_RTR_FLAG = 0x40000000
CAN_EFF_FLAG = 0x80000000
CAN_ID_MASK_EXT = 0x1FFFFFFF
CAN_ID_MASK_STD = 0x7FF


class SerialBus(BusABC):
    """
    Enable basic can communication over a serial device.

    .. note:: See :meth:`~_recv_internal` for some special semantics.

    """

    def __init__(
        self,
        channel: str,
        baudrate: int = 115200,
        timeout: float = 0.1,
        rtscts: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        :param channel:
            The serial device to open. For example "/dev/ttyS1" or
            "/dev/ttyUSB0" on Linux or "COM1" on Windows systems.

        :param baudrate:
            Baud rate of the serial device in bit/s (default 115200).

            .. warning::
                Some serial port implementations don't care about the baudrate.

        :param timeout:
            Timeout for the serial device in seconds (default 0.1).

        :param rtscts:
            turn hardware handshake (RTS/CTS) on and off

        :raises ~can.exceptions.CanInitializationError:
            If the given parameters are invalid.
        :raises ~can.exceptions.CanInterfaceNotImplementedError:
            If the serial module is not installed.
        """

        if not serial:
            raise CanInterfaceNotImplementedError("the serial module is not installed")

        if not channel:
            raise TypeError("Must specify a serial port.")

        self.channel_info = f"Serial interface: {channel}"
        self._can_protocol = CanProtocol.CAN_20

        try:
            self._ser = serial.serial_for_url(
                channel, baudrate=baudrate, timeout=timeout, rtscts=rtscts
            )
        except ValueError as error:
            raise CanInitializationError(
                "could not create the serial device"
            ) from error

        super().__init__(channel, **kwargs)

    def shutdown(self) -> None:
        """
        Close the serial interface.
        """
        super().shutdown()
        self._ser.close()

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        """
        Send a message over the serial device.

        :param msg:
            Message to send.

            .. note:: If the timestamp is a float value it will be converted
                      to an integer.

        :param timeout:
            This parameter will be ignored. The timeout value of the channel is
            used instead.

        """
        # Pack timestamp
        try:
            timestamp = struct.pack("<I", int(msg.timestamp * 1000))
        except struct.error:
            raise ValueError(f"Timestamp is out of range: {msg.timestamp}") from None

        # Pack arbitration ID
        if msg.is_extended_id:
            arbitration_id = msg.arbitration_id & CAN_ID_MASK_EXT
            arbitration_id |= CAN_EFF_FLAG
        else:
            arbitration_id = msg.arbitration_id & CAN_ID_MASK_STD

        if msg.is_error_frame:
            arbitration_id |= CAN_ERR_FLAG
        if msg.is_remote_frame:
            arbitration_id |= CAN_RTR_FLAG

        arbitration_id_bytes = struct.pack("<I", arbitration_id)

        # Assemble message
        byte_msg = bytearray()
        byte_msg.append(0xAA)
        byte_msg += timestamp
        byte_msg.append(msg.dlc)
        byte_msg += arbitration_id_bytes
        byte_msg += msg.data
        byte_msg.append(0xBB)

        # Write to serial device
        try:
            self._ser.write(byte_msg)
        except serial.PortNotOpenError as error:
            raise CanOperationError("writing to closed port") from error
        except serial.SerialTimeoutException as error:
            raise CanTimeoutError() from error

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> tuple[Optional[Message], bool]:
        """
        Read a message from the serial device.

        :param timeout:

            .. warning::
                This parameter will be ignored. The timeout value of the channel is used.

        :returns:
            Received message and :obj:`False` (because no filtering as taken place).
        """
        try:
            rx_byte = self._ser.read()
            if rx_byte and ord(rx_byte) == 0xAA:
                s = self._ser.read(4)
                timestamp = struct.unpack("<I", s)[0]
                dlc = ord(self._ser.read())
                if dlc > 8:
                    raise ValueError("received DLC may not exceed 8 bytes")

                s = self._ser.read(4)
                arbitration_id = struct.unpack("<I", s)[0]
                is_extended_id = bool(arbitration_id & CAN_EFF_FLAG)
                is_error_frame = bool(arbitration_id & CAN_ERR_FLAG)
                is_remote_frame = bool(arbitration_id & CAN_RTR_FLAG)

                if is_extended_id:
                    arbitration_id = arbitration_id & CAN_ID_MASK_EXT
                else:
                    arbitration_id = arbitration_id & CAN_ID_MASK_STD

                data = self._ser.read(dlc)

                delimiter_byte = ord(self._ser.read())
                if delimiter_byte == 0xBB:
                    # received message data okay
                    msg = Message(
                        # TODO: We are only guessing that they are milliseconds
                        timestamp=timestamp / 1000,
                        arbitration_id=arbitration_id,
                        dlc=dlc,
                        data=data,
                        is_extended_id=is_extended_id,
                        is_error_frame=is_error_frame,
                        is_remote_frame=is_remote_frame,
                    )
                    return msg, False

                else:
                    raise CanOperationError(
                        f"invalid delimiter byte while reading message: {delimiter_byte}"
                    )

            else:
                return None, False

        except serial.SerialException as error:
            raise CanOperationError("could not read from serial") from error

    def fileno(self) -> int:
        try:
            return cast("int", self._ser.fileno())
        except io.UnsupportedOperation:
            raise NotImplementedError(
                "fileno is not implemented using current CAN bus on this platform"
            ) from None
        except Exception as exception:
            raise CanOperationError("Cannot fetch fileno") from exception

    @staticmethod
    def _detect_available_configs() -> Sequence[AutoDetectedConfig]:
        configs: list[AutoDetectedConfig] = []
        if serial is None:
            return configs

        for port in serial.tools.list_ports.comports():
            configs.append({"interface": "serial", "channel": port.device})
        return configs



"""
Interface for slcan compatible interfaces (win32/linux).
"""

import io
import logging
import time
import warnings
from queue import SimpleQueue
from typing import Any, Optional, Union, cast

from can import BitTiming, BitTimingFd, BusABC, CanProtocol, Message, typechecking
from can.exceptions import (
    CanInitializationError,
    CanInterfaceNotImplementedError,
    CanOperationError,
    error_check,
)
from can.util import (
    CAN_FD_DLC,
    check_or_adjust_timing_clock,
    deprecated_args_alias,
    len2dlc,
)

logger = logging.getLogger(__name__)

try:
    import serial
except ImportError:
    logger.warning(
        "You won't be able to use the slcan can backend without "
        "the serial module installed!"
    )
    serial = None


class slcanBus(BusABC):
    """
    slcan interface
    """

    # the supported bitrates and their commands
    _BITRATES = {
        10000: "S0",
        20000: "S1",
        50000: "S2",
        100000: "S3",
        125000: "S4",
        250000: "S5",
        500000: "S6",
        750000: "S7",
        1000000: "S8",
        83300: "S9",
    }
    _DATA_BITRATES = {
        0: "",
        2000000: "Y2",
        5000000: "Y5",
    }

    _SLEEP_AFTER_SERIAL_OPEN = 2  # in seconds

    _OK = b"\r"
    _ERROR = b"\a"

    LINE_TERMINATOR = b"\r"

    @deprecated_args_alias(
        deprecation_start="4.5.0",
        deprecation_end="5.0.0",
        ttyBaudrate="tty_baudrate",
    )
    def __init__(
        self,
        channel: typechecking.ChannelStr,
        tty_baudrate: int = 115200,
        bitrate: Optional[int] = None,
        timing: Optional[Union[BitTiming, BitTimingFd]] = None,
        sleep_after_open: float = _SLEEP_AFTER_SERIAL_OPEN,
        rtscts: bool = False,
        listen_only: bool = False,
        timeout: float = 0.001,
        **kwargs: Any,
    ) -> None:
        """
        :param str channel:
            port of underlying serial or usb device (e.g. ``/dev/ttyUSB0``, ``COM8``, ...)
            Must not be empty. Can also end with ``@115200`` (or similarly) to specify the baudrate.
        :param int tty_baudrate:
            baudrate of underlying serial or usb device (Ignored if set via the ``channel`` parameter)
        :param bitrate:
            Bitrate in bit/s
        :param timing:
            Optional :class:`~can.BitTiming` instance to use for custom bit timing setting.
            If this argument is set then it overrides the bitrate and btr arguments. The
            `f_clock` value of the timing instance must be set to 8_000_000 (8MHz)
            for standard CAN.
            CAN FD and the :class:`~can.BitTimingFd` class have partial support according to the non-standard
            slcan protocol implementation in the CANABLE 2.0 firmware: currently only data rates of 2M and 5M.
        :param poll_interval:
            Poll interval in seconds when reading messages
        :param sleep_after_open:
            Time to wait in seconds after opening serial connection
        :param rtscts:
            turn hardware handshake (RTS/CTS) on and off
        :param listen_only:
            If True, open interface/channel in listen mode with ``L`` command.
            Otherwise, the (default) ``O`` command is still used. See ``open`` method.
        :param timeout:
            Timeout for the serial or usb device in seconds (default 0.001)

        :raise ValueError: if both ``bitrate`` and ``btr`` are set or the channel is invalid
        :raise CanInterfaceNotImplementedError: if the serial module is missing
        :raise CanInitializationError: if the underlying serial connection could not be established
        """
        self._listen_only = listen_only

        if serial is None:
            raise CanInterfaceNotImplementedError("The serial module is not installed")

        btr: Optional[str] = kwargs.get("btr", None)
        if btr is not None:
            warnings.warn(
                "The 'btr' argument is deprecated since python-can v4.5.0 "
                "and scheduled for removal in v5.0.0. "
                "Use the 'timing' argument instead.",
                DeprecationWarning,
                stacklevel=1,
            )

        if not channel:  # if None or empty
            raise ValueError("Must specify a serial port.")
        if "@" in channel:
            (channel, baudrate) = channel.split("@")
            tty_baudrate = int(baudrate)

        with error_check(exception_type=CanInitializationError):
            self.serialPortOrig = serial.serial_for_url(
                channel,
                baudrate=tty_baudrate,
                rtscts=rtscts,
                timeout=timeout,
            )

        self._queue: SimpleQueue[str] = SimpleQueue()
        self._buffer = bytearray()
        self._can_protocol = CanProtocol.CAN_20

        time.sleep(sleep_after_open)

        with error_check(exception_type=CanInitializationError):
            if isinstance(timing, BitTiming):
                timing = check_or_adjust_timing_clock(timing, valid_clocks=[8_000_000])
                self.set_bitrate_reg(f"{timing.btr0:02X}{timing.btr1:02X}")
            elif isinstance(timing, BitTimingFd):
                self.set_bitrate(timing.nom_bitrate, timing.data_bitrate)
            else:
                if bitrate is not None and btr is not None:
                    raise ValueError("Bitrate and btr mutually exclusive.")
                if bitrate is not None:
                    self.set_bitrate(bitrate)
                if btr is not None:
                    self.set_bitrate_reg(btr)
            self.open()

        super().__init__(channel, **kwargs)

    def set_bitrate(self, bitrate: int, data_bitrate: Optional[int] = None) -> None:
        """
        :param bitrate:
            Bitrate in bit/s
        :param data_bitrate:
            Data Bitrate in bit/s for FD frames

        :raise ValueError: if ``bitrate`` is not among the possible values
        """
        if bitrate in self._BITRATES:
            bitrate_code = self._BITRATES[bitrate]
        else:
            bitrates = ", ".join(str(k) for k in self._BITRATES.keys())
            raise ValueError(f"Invalid bitrate, choose one of {bitrates}.")

        # If data_bitrate is None, we set it to 0 which means no data bitrate
        if data_bitrate is None:
            data_bitrate = 0

        if data_bitrate in self._DATA_BITRATES:
            dbitrate_code = self._DATA_BITRATES[data_bitrate]
        else:
            dbitrates = ", ".join(str(k) for k in self._DATA_BITRATES.keys())
            raise ValueError(f"Invalid data bitrate, choose one of {dbitrates}.")

        self.close()
        self._write(bitrate_code)
        self._write(dbitrate_code)
        self.open()

    def set_bitrate_reg(self, btr: str) -> None:
        """
        :param btr:
            BTR register value to set custom can speed as a string `xxyy` where
            xx is the BTR0 value in hex and yy is the BTR1 value in hex.
        """
        self.close()
        self._write("s" + btr)
        self.open()

    def _write(self, string: str) -> None:
        with error_check("Could not write to serial device"):
            self.serialPortOrig.write(string.encode() + self.LINE_TERMINATOR)
            self.serialPortOrig.flush()

    def _read(self, timeout: Optional[float]) -> Optional[str]:
        _timeout = serial.Timeout(timeout)

        with error_check("Could not read from serial device"):
            while True:
                # Due to accessing `serialPortOrig.in_waiting` too often will reduce the performance.
                # We read the `serialPortOrig.in_waiting` only once here.
                in_waiting = self.serialPortOrig.in_waiting
                for _ in range(max(1, in_waiting)):
                    new_byte = self.serialPortOrig.read(1)
                    if new_byte:
                        self._buffer.extend(new_byte)
                    else:
                        break

                    if new_byte in (self._ERROR, self._OK):
                        string = self._buffer.decode()
                        self._buffer.clear()
                        return string

                if _timeout.expired():
                    break

            return None

    def flush(self) -> None:
        self._buffer.clear()
        with error_check("Could not flush"):
            self.serialPortOrig.reset_input_buffer()

    def open(self) -> None:
        if self._listen_only:
            self._write("L")
        else:
            self._write("O")

    def close(self) -> None:
        self._write("C")

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> tuple[Optional[Message], bool]:
        canId = None
        remote = False
        extended = False
        data = None
        isFd = False
        fdBrs = False

        if self._queue.qsize():
            string: Optional[str] = self._queue.get_nowait()
        else:
            string = self._read(timeout)

        if not string:
            pass
        elif string[0] in (
            "T",
            "x",  # x is an alternative extended message identifier for CANDapter
        ):
            # extended frame
            canId = int(string[1:9], 16)
            dlc = int(string[9])
            extended = True
            data = bytearray.fromhex(string[10 : 10 + dlc * 2])
        elif string[0] == "t":
            # normal frame
            canId = int(string[1:4], 16)
            dlc = int(string[4])
            data = bytearray.fromhex(string[5 : 5 + dlc * 2])
        elif string[0] == "r":
            # remote frame
            canId = int(string[1:4], 16)
            dlc = int(string[4])
            remote = True
        elif string[0] == "R":
            # remote extended frame
            canId = int(string[1:9], 16)
            dlc = int(string[9])
            extended = True
            remote = True
        elif string[0] == "d":
            # FD standard frame
            canId = int(string[1:4], 16)
            dlc = int(string[4], 16)
            isFd = True
            data = bytearray.fromhex(string[5 : 5 + CAN_FD_DLC[dlc] * 2])
        elif string[0] == "D":
            # FD extended frame
            canId = int(string[1:9], 16)
            dlc = int(string[9], 16)
            extended = True
            isFd = True
            data = bytearray.fromhex(string[10 : 10 + CAN_FD_DLC[dlc] * 2])
        elif string[0] == "b":
            # FD with bitrate switch
            canId = int(string[1:4], 16)
            dlc = int(string[4], 16)
            isFd = True
            fdBrs = True
            data = bytearray.fromhex(string[5 : 5 + CAN_FD_DLC[dlc] * 2])
        elif string[0] == "B":
            # FD extended with bitrate switch
            canId = int(string[1:9], 16)
            dlc = int(string[9], 16)
            extended = True
            isFd = True
            fdBrs = True
            data = bytearray.fromhex(string[10 : 10 + CAN_FD_DLC[dlc] * 2])

        if canId is not None:
            msg = Message(
                arbitration_id=canId,
                is_extended_id=extended,
                timestamp=time.time(),  # Better than nothing...
                is_remote_frame=remote,
                is_fd=isFd,
                bitrate_switch=fdBrs,
                dlc=CAN_FD_DLC[dlc],
                data=data,
            )
            return msg, False
        return None, False

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        if timeout != self.serialPortOrig.write_timeout:
            self.serialPortOrig.write_timeout = timeout
        if msg.is_remote_frame:
            if msg.is_extended_id:
                sendStr = f"R{msg.arbitration_id:08X}{msg.dlc:d}"
            else:
                sendStr = f"r{msg.arbitration_id:03X}{msg.dlc:d}"
        elif msg.is_fd:
            fd_dlc = len2dlc(msg.dlc)
            if msg.bitrate_switch:
                if msg.is_extended_id:
                    sendStr = f"B{msg.arbitration_id:08X}{fd_dlc:X}"
                else:
                    sendStr = f"b{msg.arbitration_id:03X}{fd_dlc:X}"
                sendStr += msg.data.hex().upper()
            else:
                if msg.is_extended_id:
                    sendStr = f"D{msg.arbitration_id:08X}{fd_dlc:X}"
                else:
                    sendStr = f"d{msg.arbitration_id:03X}{fd_dlc:X}"
                sendStr += msg.data.hex().upper()
        else:
            if msg.is_extended_id:
                sendStr = f"T{msg.arbitration_id:08X}{msg.dlc:d}"
            else:
                sendStr = f"t{msg.arbitration_id:03X}{msg.dlc:d}"
            sendStr += msg.data.hex().upper()
        self._write(sendStr)

    def shutdown(self) -> None:
        super().shutdown()
        self.close()
        with error_check("Could not close serial socket"):
            self.serialPortOrig.close()

    def fileno(self) -> int:
        try:
            return cast("int", self.serialPortOrig.fileno())
        except io.UnsupportedOperation:
            raise NotImplementedError(
                "fileno is not implemented using current CAN bus on this platform"
            ) from None
        except Exception as exception:
            raise CanOperationError("Cannot fetch fileno") from exception

    def get_version(
        self, timeout: Optional[float]
    ) -> tuple[Optional[int], Optional[int]]:
        """Get HW and SW version of the slcan interface.

        :param timeout:
            seconds to wait for version or None to wait indefinitely

        :returns: tuple (hw_version, sw_version)
            WHERE
            int hw_version is the hardware version or None on timeout
            int sw_version is the software version or None on timeout
        """
        _timeout = serial.Timeout(timeout)
        cmd = "V"
        self._write(cmd)

        while True:
            if string := self._read(_timeout.time_left()):
                if string[0] == cmd:
                    # convert ASCII coded version
                    hw_version = int(string[1:3])
                    sw_version = int(string[3:5])
                    return hw_version, sw_version
                else:
                    self._queue.put_nowait(string)
            if _timeout.expired():
                break
        return None, None

    def get_serial_number(self, timeout: Optional[float]) -> Optional[str]:
        """Get serial number of the slcan interface.

        :param timeout:
            seconds to wait for serial number or :obj:`None` to wait indefinitely

        :return:
            :obj:`None` on timeout or a :class:`str` object.
        """
        _timeout = serial.Timeout(timeout)
        cmd = "N"
        self._write(cmd)

        while True:
            if string := self._read(_timeout.time_left()):
                if string[0] == cmd:
                    serial_number = string[1:-1]
                    return serial_number
                else:
                    self._queue.put_nowait(string)
            if _timeout.expired():
                break
        return None



"""
Creates a bridge between two CAN buses.

This will connect to two CAN buses. Messages received on one
bus will be sent to the other bus and vice versa.
"""

import argparse
import errno
import sys
import time
from datetime import datetime
from typing import Final

from can.cli import add_bus_arguments, create_bus_from_namespace
from can.listener import RedirectReader
from can.notifier import Notifier

BRIDGE_DESCRIPTION: Final = """\
Bridge two CAN buses.

Both can buses will be connected so that messages from bus1 will be sent on
bus2 and messages from bus2 will be sent to bus1.
"""
BUS_1_PREFIX: Final = "bus1"
BUS_2_PREFIX: Final = "bus2"


def _parse_bridge_args(args: list[str]) -> argparse.Namespace:
    """Parse command line arguments for bridge script."""

    parser = argparse.ArgumentParser(description=BRIDGE_DESCRIPTION)
    add_bus_arguments(parser, prefix=BUS_1_PREFIX, group_title="Bus 1 arguments")
    add_bus_arguments(parser, prefix=BUS_2_PREFIX, group_title="Bus 2 arguments")

    # print help message when no arguments were given
    if not args:
        parser.print_help(sys.stderr)
        raise SystemExit(errno.EINVAL)

    results, _unknown_args = parser.parse_known_args(args)
    return results


def main() -> None:
    results = _parse_bridge_args(sys.argv[1:])

    with (
        create_bus_from_namespace(results, prefix=BUS_1_PREFIX) as bus1,
        create_bus_from_namespace(results, prefix=BUS_2_PREFIX) as bus2,
    ):
        reader1_to_2 = RedirectReader(bus2)
        reader2_to_1 = RedirectReader(bus1)
        with Notifier(bus1, [reader1_to_2]), Notifier(bus2, [reader2_to_1]):
            print(f"CAN Bridge (Started on {datetime.now()})")
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                pass

    print(f"CAN Bridge (Stopped on {datetime.now()})")


if __name__ == "__main__":
    main()



"""
This module implements an OS and hardware independent
virtual CAN interface for testing purposes.

Any VirtualBus instances connecting to the same channel
and reside in the same process will receive the same messages.
"""

import logging
import queue
import time
from copy import deepcopy
from random import randint
from threading import RLock
from typing import Any, Final, Optional

from can import CanOperationError
from can.bus import BusABC, CanProtocol
from can.message import Message
from can.typechecking import AutoDetectedConfig, Channel

logger = logging.getLogger(__name__)

# Channels are lists of queues, one for each connection
channels: Final[dict[Channel, list[queue.Queue[Message]]]] = {}
channels_lock: Final = RLock()


class VirtualBus(BusABC):
    """
    A virtual CAN bus using an internal message queue. It can be used for
    example for testing.

    In this interface, a channel is an arbitrary object used as
    an identifier for connected buses.

    Implements :meth:`can.BusABC._detect_available_configs`; see
    :meth:`_detect_available_configs` for how it
    behaves here.

    .. note::
        The timeout when sending a message applies to each receiver
        individually. This means that sending can block up to 5 seconds
        if a message is sent to 5 receivers with the timeout set to 1.0.

    .. warning::
        This interface guarantees reliable delivery and message ordering, but
        does *not* implement rate limiting or ID arbitration/prioritization
        under high loads. Please refer to the section
        :ref:`virtual_interfaces_doc` for more information on this and a
        comparison to alternatives.
    """

    def __init__(
        self,
        channel: Channel = "channel-0",
        receive_own_messages: bool = False,
        rx_queue_size: int = 0,
        preserve_timestamps: bool = False,
        protocol: CanProtocol = CanProtocol.CAN_20,
        **kwargs: Any,
    ) -> None:
        """
        The constructed instance has access to the bus identified by the
        channel parameter. It is able to see all messages transmitted on the
        bus by virtual instances constructed with the same channel identifier.

        :param channel: The channel identifier. This parameter can be an
            arbitrary hashable value. The bus instance will be able to see
            messages from other virtual bus instances that were created with
            the same value.
        :param receive_own_messages: If set to True, sent messages will be
            reflected back on the input queue.
        :param rx_queue_size: The size of the reception queue. The reception
            queue stores messages until they are read. If the queue reaches
            its capacity, it will start dropping the oldest messages to make
            room for new ones. If set to 0, the queue has an infinite capacity.
            Be aware that this can cause memory leaks if messages are read
            with a lower frequency than they arrive on the bus.
        :param preserve_timestamps: If set to True, messages transmitted via
            :func:`~can.BusABC.send` will keep the timestamp set in the
            :class:`~can.Message` instance. Otherwise, the timestamp value
            will be replaced with the current system time.
        :param protocol: The protocol implemented by this bus instance. The
            value does not affect the operation of the bus instance and can
            be set to an arbitrary value for testing purposes.
        :param kwargs: Additional keyword arguments passed to the parent
            constructor.
        """
        super().__init__(
            channel=channel,
            receive_own_messages=receive_own_messages,
            **kwargs,
        )

        # the channel identifier may be an arbitrary object
        self.channel_id = channel
        self._can_protocol = protocol
        self.channel_info = f"Virtual bus channel {self.channel_id}"
        self.receive_own_messages = receive_own_messages
        self.preserve_timestamps = preserve_timestamps
        self._open = True

        with channels_lock:
            # Create a new channel if one does not exist
            if self.channel_id not in channels:
                channels[self.channel_id] = []
            self.channel = channels[self.channel_id]

            self.queue: queue.Queue[Message] = queue.Queue(rx_queue_size)
            self.channel.append(self.queue)

    def _check_if_open(self) -> None:
        """Raises :exc:`~can.exceptions.CanOperationError` if the bus is not open.

        Has to be called in every method that accesses the bus.
        """
        if not self._open:
            raise CanOperationError("Cannot operate on a closed bus")

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> tuple[Optional[Message], bool]:
        self._check_if_open()
        try:
            msg = self.queue.get(block=True, timeout=timeout)
        except queue.Empty:
            return None, False
        else:
            return msg, False

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        self._check_if_open()

        timestamp = msg.timestamp if self.preserve_timestamps else time.time()
        # Add message to all listening on this channel
        all_sent = True
        for bus_queue in self.channel:
            if bus_queue is self.queue and not self.receive_own_messages:
                continue
            msg_copy = deepcopy(msg)
            msg_copy.timestamp = timestamp
            msg_copy.channel = self.channel_id
            msg_copy.is_rx = bus_queue is not self.queue
            try:
                bus_queue.put(msg_copy, block=True, timeout=timeout)
            except queue.Full:
                all_sent = False

        if not all_sent:
            raise CanOperationError("Could not send message to one or more recipients")

    def shutdown(self) -> None:
        super().shutdown()
        if self._open:
            self._open = False

            with channels_lock:
                self.channel.remove(self.queue)

                # remove if empty
                if not self.channel:
                    del channels[self.channel_id]

    @staticmethod
    def _detect_available_configs() -> list[AutoDetectedConfig]:
        """
        Returns all currently used channels as well as
        one other currently unused channel.

        .. note::

            This method will run into problems if thousands of
            autodetected buses are used at once.

        """
        with channels_lock:
            available_channels = list(channels.keys())

        # find a currently unused channel
        def get_extra() -> str:
            return f"channel-{randint(0, 9999)}"

        extra = get_extra()
        while extra in available_channels:
            extra = get_extra()

        available_channels += [extra]

        return [
            {"interface": "virtual", "channel": channel}
            for channel in available_channels
        ]



"""
This module contains handling for CSV (comma separated values) files.

TODO: CAN FD messages are not yet supported.

TODO: This module could use https://docs.python.org/2/library/csv.html#module-csv
      to allow different delimiters for writing, special escape chars to circumvent
      the base64 encoding and use csv.Sniffer to automatically deduce the delimiters
      of a CSV file.
"""

from base64 import b64decode, b64encode
from collections.abc import Generator
from typing import Any, TextIO, Union

from can.message import Message

from ..typechecking import StringPathLike
from .generic import TextIOMessageReader, TextIOMessageWriter


class CSVReader(TextIOMessageReader):
    """Iterator over CAN messages from a .csv file that was
    generated by :class:`~can.CSVWriter` or that uses the same
    format as described there. Assumes that there is a header
    and thus skips the first line.

    Any line separator is accepted.
    """

    def __init__(
        self,
        file: Union[StringPathLike, TextIO],
        **kwargs: Any,
    ) -> None:
        """
        :param file: a path-like object or as file-like object to read from
                     If this is a file-like object, is has to opened in text
                     read mode, not binary read mode.
        """
        super().__init__(file, mode="r")

    def __iter__(self) -> Generator[Message, None, None]:
        # skip the header line
        try:
            next(self.file)
        except StopIteration:
            # don't crash on a file with only a header
            return

        for line in self.file:
            timestamp, arbitration_id, extended, remote, error, dlc, data = line.split(
                ","
            )

            yield Message(
                timestamp=float(timestamp),
                is_remote_frame=(remote == "1"),
                is_extended_id=(extended == "1"),
                is_error_frame=(error == "1"),
                arbitration_id=int(arbitration_id, base=16),
                dlc=int(dlc),
                data=b64decode(data),
            )

        self.stop()


class CSVWriter(TextIOMessageWriter):
    """Writes a comma separated text file with a line for
    each message. Includes a header line.

    The columns are as follows:

    ================ ======================= ===============
    name of column   format description      example
    ================ ======================= ===============
    timestamp        decimal float           1483389946.197
    arbitration_id   hex                     0x00dadada
    extended         1 == True, 0 == False   1
    remote           1 == True, 0 == False   0
    error            1 == True, 0 == False   0
    dlc              int                     6
    data             base64 encoded          WzQyLCA5XQ==
    ================ ======================= ===============

    Each line is terminated with a platform specific line separator.
    """

    def __init__(
        self,
        file: Union[StringPathLike, TextIO],
        append: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        :param file: a path-like object or a file-like object to write to.
                     If this is a file-like object, is has to open in text
                     write mode, not binary write mode.
        :param bool append: if set to `True` messages are appended to
                            the file and no header line is written, else
                            the file is truncated and starts with a newly
                            written header line
        """
        super().__init__(file, mode="a" if append else "w")

        # Write a header row
        if not append:
            self.file.write("timestamp,arbitration_id,extended,remote,error,dlc,data\n")

    def on_message_received(self, msg: Message) -> None:
        row = ",".join(
            [
                repr(msg.timestamp),  # cannot use str() here because that is rounding
                hex(msg.arbitration_id),
                "1" if msg.is_extended_id else "0",
                "1" if msg.is_remote_frame else "0",
                "1" if msg.is_error_frame else "0",
                str(msg.dlc),
                b64encode(msg.data).decode("utf8"),
            ]
        )
        self.file.write(row)
        self.file.write("\n")



""" """

__all__ = [
    "CANLIBInitializationError",
    "CANLIBOperationError",
    "KvaserBus",
    "canlib",
    "constants",
    "get_channel_info",
    "structures",
]

from can.interfaces.kvaser.canlib import *



"""
This module contains the implementation of :class:`~can.Notifier`.
"""

import asyncio
import functools
import logging
import threading
import time
from collections.abc import Awaitable, Iterable
from contextlib import AbstractContextManager
from types import TracebackType
from typing import (
    Any,
    Callable,
    Final,
    NamedTuple,
    Optional,
    Union,
)

from can.bus import BusABC
from can.listener import Listener
from can.message import Message

logger = logging.getLogger("can.Notifier")

MessageRecipient = Union[Listener, Callable[[Message], Union[Awaitable[None], None]]]


class _BusNotifierPair(NamedTuple):
    bus: "BusABC"
    notifier: "Notifier"


class _NotifierRegistry:
    """A registry to manage the association between CAN buses and Notifiers.

    This class ensures that a bus is not added to multiple active Notifiers.
    """

    def __init__(self) -> None:
        """Initialize the registry with an empty list of bus-notifier pairs and a threading lock."""
        self.pairs: list[_BusNotifierPair] = []
        self.lock = threading.Lock()

    def register(self, bus: BusABC, notifier: "Notifier") -> None:
        """Register a bus and its associated notifier.

        Ensures that a bus is not added to multiple active :class:`~can.Notifier` instances.

        :param bus:
            The CAN bus to register.
        :param notifier:
            The :class:`~can.Notifier` instance associated with the bus.
        :raises ValueError:
            If the bus is already assigned to an active Notifier.
        """
        with self.lock:
            for pair in self.pairs:
                if bus is pair.bus and not pair.notifier.stopped:
                    raise ValueError(
                        "A bus can not be added to multiple active Notifier instances."
                    )
            self.pairs.append(_BusNotifierPair(bus, notifier))

    def unregister(self, bus: BusABC, notifier: "Notifier") -> None:
        """Unregister a bus and its associated notifier.

        Removes the bus-notifier pair from the registry.

        :param bus:
            The CAN bus to unregister.
        :param notifier:
            The :class:`~can.Notifier` instance associated with the bus.
        """
        with self.lock:
            registered_pairs_to_remove: list[_BusNotifierPair] = []
            for pair in self.pairs:
                if pair.bus is bus and pair.notifier is notifier:
                    registered_pairs_to_remove.append(pair)
            for pair in registered_pairs_to_remove:
                self.pairs.remove(pair)

    def find_instances(self, bus: BusABC) -> tuple["Notifier", ...]:
        """Find the :class:`~can.Notifier` instances associated with a given CAN bus.

        This method searches the registry for the :class:`~can.Notifier`
        that is linked to the specified bus. If the bus is found, the
        corresponding :class:`~can.Notifier` instances are returned. If the bus is not
        found in the registry, an empty tuple is returned.

        :param bus:
            The CAN bus for which to find the associated :class:`~can.Notifier` .
        :return:
            A tuple of :class:`~can.Notifier` instances associated with the given bus.
        """
        instance_list = []
        with self.lock:
            for pair in self.pairs:
                if bus is pair.bus:
                    instance_list.append(pair.notifier)
        return tuple(instance_list)


class Notifier(AbstractContextManager["Notifier"]):

    _registry: Final = _NotifierRegistry()

    def __init__(
        self,
        bus: Union[BusABC, list[BusABC]],
        listeners: Iterable[MessageRecipient],
        timeout: float = 1.0,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """Manages the distribution of :class:`~can.Message` instances to listeners.

        Supports multiple buses and listeners.

        .. Note::

            Remember to call :meth:`~can.Notifier.stop` after all messages are received as
            many listeners carry out flush operations to persist data.


        :param bus:
            A :ref:`bus` or a list of buses to consume messages from.
        :param listeners:
            An iterable of :class:`~can.Listener` or callables that receive a :class:`~can.Message`
            and return nothing.
        :param timeout:
            An optional maximum number of seconds to wait for any :class:`~can.Message`.
        :param loop:
            An :mod:`asyncio` event loop to schedule the ``listeners`` in.
        :raises ValueError:
            If a passed in *bus* is already assigned to an active :class:`~can.Notifier`.
        """
        self.listeners: list[MessageRecipient] = list(listeners)
        self._bus_list: list[BusABC] = []
        self.timeout = timeout
        self._loop = loop

        #: Exception raised in thread
        self.exception: Optional[Exception] = None

        self._stopped = False
        self._lock = threading.Lock()

        self._readers: list[Union[int, threading.Thread]] = []
        _bus_list: list[BusABC] = bus if isinstance(bus, list) else [bus]
        for each_bus in _bus_list:
            self.add_bus(each_bus)

    @property
    def bus(self) -> Union[BusABC, tuple["BusABC", ...]]:
        """Return the associated bus or a tuple of buses."""
        if len(self._bus_list) == 1:
            return self._bus_list[0]
        return tuple(self._bus_list)

    def add_bus(self, bus: BusABC) -> None:
        """Add a bus for notification.

        :param bus:
            CAN bus instance.
        :raises ValueError:
            If the *bus* is already assigned to an active :class:`~can.Notifier`.
        """
        # add bus to notifier registry
        Notifier._registry.register(bus, self)

        # add bus to internal bus list
        self._bus_list.append(bus)

        file_descriptor: int = -1
        try:
            file_descriptor = bus.fileno()
        except NotImplementedError:
            # Bus doesn't support fileno, we fall back to thread based reader
            pass

        if self._loop is not None and file_descriptor >= 0:
            # Use bus file descriptor to watch for messages
            self._loop.add_reader(file_descriptor, self._on_message_available, bus)
            self._readers.append(file_descriptor)
        else:
            reader_thread = threading.Thread(
                target=self._rx_thread,
                args=(bus,),
                name=f'{self.__class__.__qualname__} for bus "{bus.channel_info}"',
            )
            reader_thread.daemon = True
            reader_thread.start()
            self._readers.append(reader_thread)

    def stop(self, timeout: float = 5.0) -> None:
        """Stop notifying Listeners when new :class:`~can.Message` objects arrive
        and call :meth:`~can.Listener.stop` on each Listener.

        :param timeout:
            Max time in seconds to wait for receive threads to finish.
            Should be longer than timeout given at instantiation.
        """
        self._stopped = True
        end_time = time.time() + timeout
        for reader in self._readers:
            if isinstance(reader, threading.Thread):
                now = time.time()
                if now < end_time:
                    reader.join(end_time - now)
            elif self._loop:
                # reader is a file descriptor
                self._loop.remove_reader(reader)
        for listener in self.listeners:
            if hasattr(listener, "stop"):
                listener.stop()

        # remove bus from registry
        for bus in self._bus_list:
            Notifier._registry.unregister(bus, self)

    def _rx_thread(self, bus: BusABC) -> None:
        # determine message handling callable early, not inside while loop
        if self._loop:
            handle_message: Callable[[Message], Any] = functools.partial(
                self._loop.call_soon_threadsafe,
                self._on_message_received,  # type: ignore[arg-type]
            )
        else:
            handle_message = self._on_message_received

        while not self._stopped:
            try:
                if msg := bus.recv(self.timeout):
                    with self._lock:
                        handle_message(msg)
            except Exception as exc:  # pylint: disable=broad-except
                self.exception = exc
                if self._loop is not None:
                    self._loop.call_soon_threadsafe(self._on_error, exc)
                    # Raise anyway
                    raise
                elif not self._on_error(exc):
                    # If it was not handled, raise the exception here
                    raise
                else:
                    # It was handled, so only log it
                    logger.debug("suppressed exception: %s", exc)

    def _on_message_available(self, bus: BusABC) -> None:
        if msg := bus.recv(0):
            self._on_message_received(msg)

    def _on_message_received(self, msg: Message) -> None:
        for callback in self.listeners:
            res = callback(msg)
            if res and self._loop and asyncio.iscoroutine(res):
                # Schedule coroutine
                self._loop.create_task(res)

    def _on_error(self, exc: Exception) -> bool:
        """Calls ``on_error()`` for all listeners if they implement it.

        :returns: ``True`` if at least one error handler was called.
        """
        was_handled = False

        for listener in self.listeners:
            if hasattr(listener, "on_error"):
                try:
                    listener.on_error(exc)
                except NotImplementedError:
                    pass
                else:
                    was_handled = True

        return was_handled

    def add_listener(self, listener: MessageRecipient) -> None:
        """Add new Listener to the notification list.
        If it is already present, it will be called two times
        each time a message arrives.

        :param listener: Listener to be added to the list to be notified
        """
        self.listeners.append(listener)

    def remove_listener(self, listener: MessageRecipient) -> None:
        """Remove a listener from the notification list. This method
        throws an exception if the given listener is not part of the
        stored listeners.

        :param listener: Listener to be removed from the list to be notified
        :raises ValueError: if `listener` was never added to this notifier
        """
        self.listeners.remove(listener)

    @property
    def stopped(self) -> bool:
        """Return ``True``, if Notifier was properly shut down with :meth:`~can.Notifier.stop`."""
        return self._stopped

    @staticmethod
    def find_instances(bus: BusABC) -> tuple["Notifier", ...]:
        """Find :class:`~can.Notifier` instances associated with a given CAN bus.

        This method searches the registry for the :class:`~can.Notifier`
        that is linked to the specified bus. If the bus is found, the
        corresponding :class:`~can.Notifier` instances are returned. If the bus is not
        found in the registry, an empty tuple is returned.

        :param bus:
            The CAN bus for which to find the associated :class:`~can.Notifier` .
        :return:
            A tuple of :class:`~can.Notifier` instances associated with the given bus.
        """
        return Notifier._registry.find_instances(bus)

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if not self._stopped:
            self.stop()



"""
Interface to socketcand
see https://github.com/linux-can/socketcand

Authors: Marvin Seiler, Gerrit Telkamp

Copyright (C) 2021  DOMOLOGIC GmbH
http://www.domologic.de
"""

import logging
import os
import select
import socket
import time
import traceback
import urllib.parse as urlparselib
import xml.etree.ElementTree as ET
from collections import deque

import can

log = logging.getLogger(__name__)

DEFAULT_SOCKETCAND_DISCOVERY_ADDRESS = ""
DEFAULT_SOCKETCAND_DISCOVERY_PORT = 42000


def detect_beacon(timeout_ms: int = 3100) -> list[can.typechecking.AutoDetectedConfig]:
    """
    Detects socketcand servers

    This is what :meth:`can.detect_available_configs` ends up calling to search
    for available socketcand servers with a default timeout of 3100ms
    (socketcand sends a beacon packet every 3000ms).

    Using this method directly allows for adjusting the timeout. Extending
    the timeout beyond the default time period could be useful if UDP
    packet loss is a concern.

    :param timeout_ms:
        Timeout in milliseconds to wait for socketcand beacon packets

    :return:
        See :meth:`~can.detect_available_configs`
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.bind(
            (DEFAULT_SOCKETCAND_DISCOVERY_ADDRESS, DEFAULT_SOCKETCAND_DISCOVERY_PORT)
        )
        log.info(
            "Listening on for socketcand UDP advertisement on %s:%s",
            DEFAULT_SOCKETCAND_DISCOVERY_ADDRESS,
            DEFAULT_SOCKETCAND_DISCOVERY_PORT,
        )

        now = time.time() * 1000
        end_time = now + timeout_ms
        while (time.time() * 1000) < end_time:
            try:
                # get all sockets that are ready (can be a list with a single value
                # being self.socket or an empty list if self.socket is not ready)
                ready_receive_sockets, _, _ = select.select([sock], [], [], 1)

                if not ready_receive_sockets:
                    log.debug("No advertisement received")
                    continue

                msg = sock.recv(1024).decode("utf-8")
                root = ET.fromstring(msg)
                if root.tag != "CANBeacon":
                    log.debug("Unexpected message received over UDP")
                    continue

                det_devs = []
                det_host = None
                det_port = None
                for child in root:
                    if child.tag == "Bus":
                        bus_name = child.attrib["name"]
                        det_devs.append(bus_name)
                    elif child.tag == "URL":
                        url = urlparselib.urlparse(child.text)
                        det_host = url.hostname
                        det_port = url.port

                if not det_devs:
                    log.debug(
                        "Got advertisement, but no SocketCAN devices advertised by socketcand"
                    )
                    continue

                if (det_host is None) or (det_port is None):
                    det_host = None
                    det_port = None
                    log.debug(
                        "Got advertisement, but no SocketCAN URL advertised by socketcand"
                    )
                    continue

                log.info(f"Found SocketCAN devices: {det_devs}")
                return [
                    {
                        "interface": "socketcand",
                        "host": det_host,
                        "port": det_port,
                        "channel": channel,
                    }
                    for channel in det_devs
                ]

            except ET.ParseError:
                log.debug("Unexpected message received over UDP")
                continue

            except Exception as exc:
                # something bad happened (e.g. the interface went down)
                log.error(f"Failed to detect beacon: {exc}  {traceback.format_exc()}")
                raise OSError(
                    f"Failed to detect beacon: {exc} {traceback.format_exc()}"
                ) from exc

        return []


def convert_ascii_message_to_can_message(ascii_msg: str) -> can.Message:
    if not ascii_msg.endswith(" >"):
        log.warning(f"Missing ending character in ascii message: {ascii_msg}")
        return None

    if ascii_msg.startswith("< frame "):
        # frame_string = ascii_msg.removeprefix("< frame ").removesuffix(" >")
        frame_string = ascii_msg[8:-2]
        parts = frame_string.split(" ", 3)
        can_id, timestamp = int(parts[0], 16), float(parts[1])
        is_ext = len(parts[0]) != 3

        data = bytearray.fromhex(parts[2])
        can_dlc = len(data)
        can_message = can.Message(
            timestamp=timestamp,
            arbitration_id=can_id,
            data=data,
            dlc=can_dlc,
            is_extended_id=is_ext,
            is_rx=True,
        )
        return can_message

    if ascii_msg.startswith("< error "):
        frame_string = ascii_msg[8:-2]
        parts = frame_string.split(" ", 3)
        can_id, timestamp = int(parts[0], 16), float(parts[1])
        is_ext = len(parts[0]) != 3

        # socketcand sends no data in the error message so we don't have information
        # about the error details, therefore the can frame is created with one
        # data byte set to zero
        data = bytearray([0])
        can_dlc = len(data)
        can_message = can.Message(
            timestamp=timestamp,
            arbitration_id=can_id & 0x1FFFFFFF,
            is_error_frame=True,
            data=data,
            dlc=can_dlc,
            is_extended_id=True,
            is_rx=True,
        )
        return can_message

    log.warning(f"Could not parse ascii message: {ascii_msg}")
    return None


def convert_can_message_to_ascii_message(can_message: can.Message) -> str:
    # Note: socketcan bus adds extended flag, remote_frame_flag & error_flag to id
    # not sure if that is necessary here
    can_id = can_message.arbitration_id
    if can_message.is_extended_id:
        can_id_string = f"{(can_id&0x1FFFFFFF):08X}"
    else:
        can_id_string = f"{(can_id&0x7FF):03X}"
    # Note: seems like we cannot add CANFD_BRS (bitrate_switch) and CANFD_ESI (error_state_indicator) flags
    data = can_message.data
    length = can_message.dlc
    bytes_string = " ".join(f"{x:x}" for x in data[0:length])
    return f"< send {can_id_string} {length:X} {bytes_string} >"


def connect_to_server(s, host, port):
    timeout_ms = 10000
    now = time.time() * 1000
    end_time = now + timeout_ms
    while now < end_time:
        try:
            s.connect((host, port))
            return
        except Exception as e:
            log.warning(f"Failed to connect to server: {type(e)} Message: {e}")
            now = time.time() * 1000
    raise TimeoutError(
        f"connect_to_server: Failed to connect server for {timeout_ms} ms"
    )


class SocketCanDaemonBus(can.BusABC):
    def __init__(self, channel, host, port, tcp_tune=False, can_filters=None, **kwargs):
        """Connects to a CAN bus served by socketcand.

        It implements :meth:`can.BusABC._detect_available_configs` to search for
        available interfaces.

        It will attempt to connect to the server for up to 10s, after which a
        TimeoutError exception will be thrown.

        If the handshake with the socketcand server fails, a CanError exception
        is thrown.

        :param channel:
            The can interface name served by socketcand.
            An example channel would be 'vcan0' or 'can0'.
        :param host:
            The host address of the socketcand server.
        :param port:
            The port of the socketcand server.
        :param tcp_tune:
            This tunes the TCP socket for low latency (TCP_NODELAY, and
            TCP_QUICKACK).
            This option is not available under windows.
        :param can_filters:
            See :meth:`can.BusABC.set_filters`.
        """
        self.__host = host
        self.__port = port

        self.__tcp_tune = tcp_tune
        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.__tcp_tune:
            if os.name == "nt":
                self.__tcp_tune = False
                log.warning("'tcp_tune' not available in Windows. Setting to False")
            else:
                self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        self.__message_buffer = deque()
        self.__receive_buffer = ""  # i know string is not the most efficient here
        self.channel = channel
        self.channel_info = f"socketcand on {channel}@{host}:{port}"
        connect_to_server(self.__socket, self.__host, self.__port)
        self._expect_msg("< hi >")

        log.info(
            f"SocketCanDaemonBus: connected with address {self.__socket.getsockname()}"
        )
        self._tcp_send(f"< open {channel} >")
        self._expect_msg("< ok >")
        self._tcp_send("< rawmode >")
        self._expect_msg("< ok >")
        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

    def _recv_internal(self, timeout):
        if len(self.__message_buffer) != 0:
            can_message = self.__message_buffer.popleft()
            return can_message, False

        try:
            # get all sockets that are ready (can be a list with a single value
            # being self.socket or an empty list if self.socket is not ready)
            ready_receive_sockets, _, _ = select.select(
                [self.__socket], [], [], timeout
            )
        except OSError as exc:
            # something bad happened (e.g. the interface went down)
            log.error(f"Failed to receive: {exc}")
            raise can.CanError(f"Failed to receive: {exc}") from exc

        try:
            if not ready_receive_sockets:
                # socket wasn't readable or timeout occurred
                log.debug("Socket not ready")
                return None, False

            ascii_msg = self.__socket.recv(1024).decode(
                "ascii"
            )  # may contain multiple messages
            if self.__tcp_tune:
                self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)
            self.__receive_buffer += ascii_msg
            log.debug(f"Received Ascii Message: {ascii_msg}")
            buffer_view = self.__receive_buffer
            chars_processed_successfully = 0
            while True:
                if len(buffer_view) == 0:
                    break

                start = buffer_view.find("<")
                if start == -1:
                    log.warning(
                        f"Bad data: No opening < found => discarding entire buffer '{buffer_view}'"
                    )
                    chars_processed_successfully = len(self.__receive_buffer)
                    break
                end = buffer_view.find(">")
                if end == -1:
                    log.warning("Got incomplete message => waiting for more data")
                    if len(buffer_view) > 200:
                        log.warning(
                            "Incomplete message exceeds 200 chars => Discarding"
                        )
                        chars_processed_successfully = len(self.__receive_buffer)
                    break
                chars_processed_successfully += end + 1
                single_message = buffer_view[start : end + 1]
                parsed_can_message = convert_ascii_message_to_can_message(
                    single_message
                )
                if parsed_can_message is None:
                    log.warning(f"Invalid Frame: {single_message}")
                else:
                    parsed_can_message.channel = self.channel
                    self.__message_buffer.append(parsed_can_message)
                buffer_view = buffer_view[end + 1 :]

            self.__receive_buffer = self.__receive_buffer[chars_processed_successfully:]
            can_message = (
                None
                if len(self.__message_buffer) == 0
                else self.__message_buffer.popleft()
            )
            return can_message, False

        except Exception as exc:
            log.error(f"Failed to receive: {exc}  {traceback.format_exc()}")
            raise can.CanError(
                f"Failed to receive: {exc}  {traceback.format_exc()}"
            ) from exc

    def _tcp_send(self, msg: str):
        log.debug(f"Sending TCP Message: '{msg}'")
        self.__socket.sendall(msg.encode("ascii"))
        if self.__tcp_tune:
            self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)

    def _expect_msg(self, msg):
        ascii_msg = self.__socket.recv(256).decode("ascii")
        if self.__tcp_tune:
            self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)
        if not ascii_msg == msg:
            raise can.CanError(f"Expected '{msg}' got: '{ascii_msg}'")

    def send(self, msg, timeout=None):
        """Transmit a message to the CAN bus.

        :param msg: A message object.
        :param timeout: Ignored
        """
        ascii_msg = convert_can_message_to_ascii_message(msg)
        self._tcp_send(ascii_msg)

    def shutdown(self):
        """Stops all active periodic tasks and closes the socket."""
        super().shutdown()
        self.__socket.close()

    @staticmethod
    def _detect_available_configs() -> list[can.typechecking.AutoDetectedConfig]:
        try:
            return detect_beacon()
        except Exception as e:
            log.warning(f"Could not detect socketcand beacon: {e}")
            return []



"""
This Listener simply prints to stdout / the terminal or a file.
"""

import logging
import sys
from io import TextIOWrapper
from typing import Any, TextIO, Union

from ..message import Message
from ..typechecking import StringPathLike
from .generic import TextIOMessageWriter

log = logging.getLogger("can.io.printer")


class Printer(TextIOMessageWriter):
    """
    The Printer class is a subclass of :class:`~can.Listener` which simply prints
    any messages it receives to the terminal (stdout). A message is turned into a
    string using :meth:`~can.Message.__str__`.

    :attr write_to_file: `True` if this instance prints to a file instead of
                         standard out
    """

    def __init__(
        self,
        file: Union[StringPathLike, TextIO, TextIOWrapper] = sys.stdout,
        append: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        :param file: An optional path-like object or a file-like object to "print"
                     to instead of writing to standard out (stdout).
                     If this is a file-like object, it has to be opened in text
                     write mode, not binary write mode.
        :param append: If set to `True` messages, are appended to the file,
                       else the file is truncated
        """
        super().__init__(file, mode="a" if append else "w")

    def on_message_received(self, msg: Message) -> None:
        self.file.write(str(msg) + "\n")

    def file_size(self) -> int:
        """Return an estimate of the current file size in bytes."""
        if self.file is not sys.stdout:
            return self.file.tell()
        return 0

    def stop(self) -> None:
        if self.file is not sys.stdout:
            super().stop()



"""
This module contains the generic :class:`LogReader` as
well as :class:`MessageSync` which plays back messages
in the recorded order and time intervals.
"""

import gzip
import pathlib
import time
from collections.abc import Generator, Iterable
from typing import (
    Any,
    Final,
)

from .._entry_points import read_entry_points
from ..message import Message
from ..typechecking import StringPathLike
from .asc import ASCReader
from .blf import BLFReader
from .canutils import CanutilsLogReader
from .csv import CSVReader
from .generic import BinaryIOMessageReader, MessageReader, TextIOMessageReader
from .mf4 import MF4Reader
from .sqlite import SqliteReader
from .trc import TRCReader

#: A map of file suffixes to their corresponding
#: :class:`can.io.generic.MessageReader` class
MESSAGE_READERS: Final[dict[str, type[MessageReader]]] = {
    ".asc": ASCReader,
    ".blf": BLFReader,
    ".csv": CSVReader,
    ".db": SqliteReader,
    ".log": CanutilsLogReader,
    ".mf4": MF4Reader,
    ".trc": TRCReader,
}


def _update_reader_plugins() -> None:
    """Update available message reader plugins from entry points."""
    for entry_point in read_entry_points("can.io.message_reader"):
        if entry_point.key in MESSAGE_READERS:
            continue

        reader_class = entry_point.load()
        if issubclass(reader_class, MessageReader):
            MESSAGE_READERS[entry_point.key] = reader_class


def _get_logger_for_suffix(suffix: str) -> type[MessageReader]:
    """Find MessageReader class for given suffix."""
    try:
        return MESSAGE_READERS[suffix]
    except KeyError:
        raise ValueError(f'No read support for unknown log format "{suffix}"') from None


def _decompress(filename: StringPathLike, **kwargs: Any) -> MessageReader:
    """
    Return the suffix and io object of the decompressed file.
    """
    suffixes = pathlib.Path(filename).suffixes
    if len(suffixes) != 2:
        raise ValueError(
            f"No read support for unknown log format \"{''.join(suffixes)}\""
        )

    real_suffix = suffixes[-2].lower()
    reader_type = _get_logger_for_suffix(real_suffix)

    if issubclass(reader_type, TextIOMessageReader):
        return reader_type(gzip.open(filename, mode="rt"), **kwargs)
    elif issubclass(reader_type, BinaryIOMessageReader):
        return reader_type(gzip.open(filename, mode="rb"), **kwargs)

    raise ValueError(f"No read support for unknown log format \"{''.join(suffixes)}\"")


def LogReader(filename: StringPathLike, **kwargs: Any) -> MessageReader:  # noqa: N802
    """Find and return the appropriate :class:`~can.io.generic.MessageReader` instance
    for a given file suffix.

    The format is determined from the file suffix which can be one of:
      * .asc :class:`can.ASCReader`
      * .blf :class:`can.BLFReader`
      * .csv :class:`can.CSVReader`
      * .db :class:`can.SqliteReader`
      * .log :class:`can.CanutilsLogReader`
      * .mf4 :class:`can.MF4Reader`
        (optional, depends on `asammdf <https://github.com/danielhrisca/asammdf>`_)
      * .trc :class:`can.TRCReader`

    Gzip compressed files can be used as long as the original
    files suffix is one of the above (e.g. filename.asc.gz).


    Exposes a simple iterator interface, to use simply::

        for msg in can.LogReader("some/path/to/my_file.log"):
            print(msg)

    :param filename:
        the filename/path of the file to read from
    :raises ValueError:
        if the filename's suffix is of an unknown file type

    .. note::
        There are no time delays, if you want to reproduce the measured
        delays between messages look at the :class:`can.MessageSync` class.

    .. note::
        This function itself is just a dispatcher, and any positional and keyword
        arguments are passed on to the returned instance.
    """

    _update_reader_plugins()

    suffix = pathlib.PurePath(filename).suffix.lower()
    if suffix == ".gz":
        return _decompress(filename)

    reader_type = _get_logger_for_suffix(suffix)
    return reader_type(file=filename, **kwargs)


class MessageSync:
    """
    Used to iterate over some given messages in the recorded time.
    """

    def __init__(
        self,
        messages: Iterable[Message],
        timestamps: bool = True,
        gap: float = 0.0001,
        skip: float = 60.0,
    ) -> None:
        """Creates an new **MessageSync** instance.

        :param messages: An iterable of :class:`can.Message` instances.
        :param timestamps: Use the messages' timestamps. If False, uses the *gap* parameter
                           as the time between messages.
        :param gap: Minimum time between sent messages in seconds
        :param skip: Skip periods of inactivity greater than this (in seconds).

        Example::

            import can

            with can.LogReader("my_logfile.asc") as reader, can.Bus(interface="virtual") as bus:
                for msg in can.MessageSync(messages=reader):
                    print(msg)
                    bus.send(msg)

        """
        self.raw_messages = messages
        self.timestamps = timestamps
        self.gap = gap
        self.skip = skip

    def __iter__(self) -> Generator[Message, None, None]:
        t_wakeup = playback_start_time = time.perf_counter()
        recorded_start_time = None
        t_skipped = 0.0

        for message in self.raw_messages:
            # Work out the correct wait time
            if self.timestamps:
                if recorded_start_time is None:
                    recorded_start_time = message.timestamp

                t_wakeup = playback_start_time + (
                    message.timestamp - t_skipped - recorded_start_time
                )
            else:
                t_wakeup += self.gap

            sleep_period = t_wakeup - time.perf_counter()

            if self.skip and sleep_period > self.skip:
                t_skipped += sleep_period - self.skip
                sleep_period = self.skip

            if sleep_period > 1e-4:
                time.sleep(sleep_period)

            yield message
